Package: stringfish
Title: Alt String Implementation
Version: 0.19.2
Date: 2026-08-04
Authors@R: c(
    person("Travers", "Ching", email = "traversc@gmail.com", role = c("aut", "cre", "cph")),
    person("Phillip", "Hazel", role = c("ctb"), comment = "Bundled PCRE2 code"),
    person("Zoltan", "Herczeg", role = c("ctb", "cph"), comment = "Bundled PCRE2 code"),
    person("University of Cambridge", role = c("cph"), comment = "Bundled PCRE2 code"),
    person("Tilera Corporation", role = c("cph"), comment = "Stack-less Just-In-Time compiler bundled with PCRE2"),
    person("Yann", "Collet", role = c("ctb", "cph"), comment = "Yann Collet is the author of the bundled xxHash code"),
    person("Martin", "Leitner-Ankerl", role = c("ctb", "cph"), comment = "Bundled ankerl::unordered_dense code"))
Maintainer: Travers Ching <traversc@gmail.com>
Description: Provides an extendable, performant and multithreaded 'alt-string' implementation backed by 'C++' vectors and strings.
License: GPL-3
Biarch: true
Encoding: UTF-8
Depends: R (>= 3.6.0)
SystemRequirements: GNU make, C++17
LinkingTo: Rcpp (>= 0.12.18.3), RcppParallel (>= 6.1.1)
Imports: Rcpp, RcppParallel
Suggests: knitr, rmarkdown
VignetteBuilder: knitr
Copyright: Copyright for the bundled 'PCRE2' library is held by
        University of Cambridge, Zoltan Herczeg and Tilera Corporation
        (Stack-less Just-In-Time compiler); Copyright for the bundled
        'xxHash' code is held by Yann Collet; Copyright for the bundled
        'ankerl::unordered_dense' code is held by Martin
        Leitner-Ankerl.
URL: https://github.com/traversc/stringfish
BugReports: https://github.com/traversc/stringfish/issues
Config/roxygen2/version: 8.0.0
NeedsCompilation: yes
Packaged: 2026-08-04 08:23:40 UTC; ted
Author: Travers Ching [aut, cre, cph],
  Phillip Hazel [ctb] (Bundled PCRE2 code),
  Zoltan Herczeg [ctb, cph] (Bundled PCRE2 code),
  University of Cambridge [cph] (Bundled PCRE2 code),
  Tilera Corporation [cph] (Stack-less Just-In-Time compiler bundled with
    PCRE2),
  Yann Collet [ctb, cph] (Yann Collet is the author of the bundled xxHash
    code),
  Martin Leitner-Ankerl [ctb, cph] (Bundled ankerl::unordered_dense code)
Repository: CRAN
Date/Publication: 2026-08-04 10:30:17 UTC
Built: R 4.6.1; aarch64-apple-darwin23; 2026-08-04 13:21:48 UTC; unix
Archs: stringfish.so.dSYM

importFrom(Rcpp,sourceCpp)
importFrom(RcppParallel, RcppParallelLibs)
useDynLib(stringfish, .registration=TRUE)
export(
"materialize",
"get_string_type",
"sf_vector",
"sf_vector_create",
"slice_store_create",
"slice_store_create_with_size",
"convert_to_sf",
"convert_to_sf_vector",
"convert_to_slice_store",
"sf_assign",
"random_strings",
"string_identical",
"sf_iconv",
"sf_nchar",
"sf_substr",
"sf_paste",
"sf_collapse",
"sf_readLines",
"sf_writeLines",
"sf_grepl",
"sf_gsub",
"sf_toupper",
"sf_tolower",
"sf_starts",
"sf_ends",
"sf_trim",
"sf_split",
"sf_match",
"sf_equals",
"sf_compare",
"sf_concat",
"sfc")

stringfish
================

[![R-CMD-check](https://github.com/traversc/stringfish/workflows/R-CMD-check/badge.svg)](https://github.com/traversc/stringfish/actions)
[![CRAN-Status-Badge](https://www.r-pkg.org/badges/version/stringfish)](https://cran.r-project.org/package=stringfish)
[![CRAN-Downloads-Badge](https://cranlogs.r-pkg.org/badges/stringfish)](https://cran.r-project.org/package=stringfish)
[![CRAN-Downloads-Total-Badge](https://cranlogs.r-pkg.org/badges/grand-total/stringfish)](https://cran.r-project.org/package=stringfish)

`stringfish` is a framework for string and sequence operations using the
ALTREP system (introduced in R 3.5) as a way to represent R objects
using custom memory layout.

This package has two primary goals:

- Provide R users a way to speed up common string operations compared to
  base R (see benchmarks below)
- Create a common ALTREP framework that can be used by other packages
  with full interoperability

`stringfish` currently provides two ALTREP backends with the same
semantics: `sf_vec`, a simple vector of string objects, and
`slice_store`, which stores strings within large contiguous blocks of
memory. They make different storage tradeoffs, but the same `stringfish`
operations work across both.

For text data, `stringfish` is intentionally UTF-8-centric outside of
explicit byte mode, so conversions, comparisons, and ALTREP views stay
consistent across normal R vectors and both backends.

## Installation

``` r
install.packages("stringfish", type="source", configure.args="--with-simd=AVX2")
```

## Benchmark

The simplest way to show the utility of the ALTREP framework is through
a quick benchmark comparing `stringfish` and base R.

<img src="vignettes/bench_v3.png" title="bench_v3" width="576" />

On favorable workloads, some functions in `stringfish` can be more than
an order of magnitude faster than vectorized base R operations, and
built-in multithreading can widen that gap further. On large text
datasets, this can turn minutes of computation into seconds.

## Currently implemented functions

A list of implemented `stringfish` functions and analogous base R
functions:

- `sf_iconv` (`iconv`)
- `sf_nchar` (`nchar`)
- `sf_substr` (`substr`)
- `sf_paste` (`paste0`)
- `sf_collapse` (`paste0`)
- `sf_readLines` (`readLines`)
- `sf_writeLines` (`writeLines`)
- `sf_grepl` (`grepl`)
- `sf_gsub` (`gsub`)
- `sf_toupper` (`toupper`)
- `sf_tolower` (`tolower`)
- `sf_starts` (`startsWith`)
- `sf_ends` (`endsWith`)
- `sf_trim` (`trimws`)
- `sf_split` (`strsplit`)
- `sf_match` (`match` for strings only)
- `sf_compare`/`sf_equals` (`==`, ALTREP-aware semantic string equality)
- `sf_concat`/`sfc` (`c`)

Utility functions:

- `sf_vector_create` – creates a new empty `sf_vec`-backed stringfish
  vector
- `sf_vector` – backwards-compatible alias for `sf_vector_create`
- `slice_store_create` – creates a new empty `slice_store`-backed
  stringfish vector
- `slice_store_create_with_size` – creates a `slice_store`-backed
  stringfish vector with an explicit initial slice size
- `sf_assign` – assign strings into a `stringfish` vector in place (like
  `x[i] <- "mystring"`)
- `convert_to_sf_vector` – converts a character vector to a `stringfish`
  vector
- `convert_to_slice_store` – converts a character vector to a
  `stringfish` slice store
- `get_string_type` – determines string type (whether ALTREP or normal)
- `materialize` – converts any ALTREP object into a normal R object
- `random_strings` – creates random strings as either a `stringfish` or
  normal R vector
- `string_identical` – compares strings either semantically or exactly
  across encodings

In addition, many R operations in base R and other packages are already
ALTREP-aware (i.e. they don’t cause materialization). Functions that
subset or index into string vectors generally do not materialize.

- `sample`
- `head`
- `tail`
- `[` – e.g. `x[20:30]`
- various tidyverse filters and operations
- Etc.

`stringfish` functions are not intended to exactly replicate their base
R analogues. One difference is that `subject` parameters are always the
first argument, which is easier to use with pipes. E.g.,
`gsub(pattern, replacement, subject)` becomes
`sf_gsub(subject, pattern, replacement)`.

## Extensibility

`stringfish` as a framework is intended to be easily extensible.
Stringfish vectors can be worked into `Rcpp` scripts or even into other
packages. The example below creates an `sf_vec`-backed output because it
is simple and direct, but the same indexing semantics work across both
backends.

Below is a detailed `Rcpp` script that creates a function to alternate
upper and lower case of strings.

``` c
// [[Rcpp::depends(stringfish)]]
#include <Rcpp.h>
#include "sf_external.h"
using namespace Rcpp;

// [[Rcpp::export]]
SEXP sf_alternate_case(SEXP x) {
  // Iterate through a character vector using the RStringIndexer class
  // If the input vector x is a stringfish character vector it will do so without materialization
  RStringIndexer r(x);
  size_t len = r.size();
  
  // Create an output stringfish vector
  // Like all R objects, it must be protected from garbage collection
  SEXP output = PROTECT(sf_vector_create(len));
  
  // Obtain a reference to the underlying output data
  sf_vec_data & output_data = sf_vec_data_ref(output);
  
  // You can use range based for loop via an iterator class that returns RStringIndexer::rstring_info e
  // rstring info is a struct containing const char * ptr, int len, and an encoding flag
  // ptr should be treated as a byte pointer plus length, not as a null-terminated C string
  // a NA string is represented by a nullptr
  // Alternatively, access the data via the function r.getCharLenCE(i)
  size_t i = 0;
  for(auto e : r) {
    // check if string is NA and go to next if it is
    if(e.ptr == nullptr) {
      i++; // increment output index
      continue;
    }
    // Create a temporary output string and process the results.
    // This example intentionally toggles ASCII letters only.
    std::string temp(e.len, '\0');
    bool case_switch = false;
    for(int j=0; j<e.len; j++) {
      if((e.ptr[j] >= 65) && (e.ptr[j] <= 90)) { // char j is upper case
        if((case_switch = !case_switch)) { // check if we should convert to lower case
          temp[j] = e.ptr[j] + 32;
          continue;
        }
      } else if((e.ptr[j] >= 97) && (e.ptr[j] <= 122)) { // char j is lower case
        if(!(case_switch = !case_switch)) { // check if we should convert to upper case
          temp[j] = e.ptr[j] - 32;
          continue;
        }
      } else if(e.ptr[j] == 32) {
        case_switch = false;
      }
      temp[j] = e.ptr[j];
    }
    
    // Create a new vector element sfstring and insert the processed string into the stringfish vector
    // sfstring has three constructors, 1) taking a std::string and encoding, 
    // 2) a char pointer and encoding, or 3) a CHARSXP object (e.g. sfstring(NA_STRING))
    output_data[i] = sfstring(temp, e.enc);
    i++; // increment output index
  }
  // Finally, call unprotect and return result
  UNPROTECT(1);
  return output;
}
```

Example function call:

``` r
sf_alternate_case("hello world") 
[1] "hElLo wOrLd"
```

## ----setup, echo=FALSE--------------------------------------------------------
IS_GITHUB <- Sys.getenv("IS_GITHUB") != ""

## ----results='asis', echo=FALSE, eval=IS_GITHUB-------------------------------
# cat('
# [![R-CMD-check](https://github.com/traversc/stringfish/workflows/R-CMD-check/badge.svg)](https://github.com/traversc/stringfish/actions)
# [![CRAN-Status-Badge](https://www.r-pkg.org/badges/version/stringfish)](https://cran.r-project.org/package=stringfish)
# [![CRAN-Downloads-Badge](https://cranlogs.r-pkg.org/badges/stringfish)](https://cran.r-project.org/package=stringfish)
# [![CRAN-Downloads-Total-Badge](https://cranlogs.r-pkg.org/badges/grand-total/stringfish)](https://cran.r-project.org/package=stringfish)
# ')

## ----eval=FALSE---------------------------------------------------------------
# install.packages("stringfish", type="source", configure.args="--with-simd=AVX2")

## ----echo=FALSE, results='asis'-----------------------------------------------
if(IS_GITHUB) {
  cat('![](vignettes/bench_v3.png "bench_v3"){width=576px}')
} else {
  cat('![](bench_v3.png "bench_v3"){width=576px}')
}

## ----eval=FALSE---------------------------------------------------------------
# sf_alternate_case("hello world")
# [1] "hElLo wOrLd"

suppressPackageStartupMessages(library(stringfish))

regression_external_source <- '
// [[Rcpp::depends(stringfish)]]
// [[Rcpp::plugins(cpp17)]]
#include <Rcpp.h>
#include "sf_external.h"
using namespace Rcpp;

// [[Rcpp::export]]
List sf_external_regressions(SEXP x, int nthreads = 1) {
  return List::create(
    _["nchar"] = sf_nchar(x, "chars", nthreads),
    _["compare"] = sf_compare(x, x, nthreads),
    _["substr"] = sf_substr(x, IntegerVector::create(1), IntegerVector::create(1), nthreads),
    _["random"] = random_strings(4, IntegerVector::create(3), "ab", "normal")
  );
}

// [[Rcpp::export]]
IntegerVector sf_external_encoding_regressions(SEXP x) {
  if(TYPEOF(x) != STRSXP || Rf_xlength(x) != 1) {
    stop("x must be a character vector of length 1");
  }
  const char native_utf8[] = "\\xC3\\xA9";
  RStringIndexer direct_indexer(x);

  SEXP vec_bytes = PROTECT(sf_vector(1));
  sf_vec_data_ref(vec_bytes)[0] = sfstring("abc", 3, CE_BYTES);
  RStringIndexer vec_bytes_indexer(vec_bytes);

  SEXP store_bytes = PROTECT(slice_store_create(1));
  slice_store_data_ref(store_bytes).assign(0, "abc", 3, CE_BYTES);
  RStringIndexer store_bytes_indexer(store_bytes);

  SEXP vec_native = PROTECT(sf_vector(1));
  sf_vec_data_ref(vec_native)[0] = sfstring(native_utf8, 2, CE_NATIVE);
  RStringIndexer vec_native_indexer(vec_native);

  SEXP store_native = PROTECT(slice_store_create(1));
  slice_store_data_ref(store_native).assign(0, native_utf8, 2, CE_NATIVE);
  RStringIndexer store_native_indexer(store_native);

  IntegerVector out = IntegerVector::create(
    _["reinterpret_ascii_bytes"] = static_cast<int>(reinterpret_input_encoding("abc", 3, CE_BYTES)),
    _["sfstring_ascii_bytes"] = static_cast<int>(sfstring("abc", 3, CE_BYTES).encoding),
    _["sf_vec_ascii_bytes"] = static_cast<int>(vec_bytes_indexer.getCharLenCE(0).enc),
    _["slice_store_ascii_bytes"] = static_cast<int>(store_bytes_indexer.getCharLenCE(0).enc),
    _["reinterpret_utf8_native"] = static_cast<int>(reinterpret_input_encoding(native_utf8, 2, CE_NATIVE)),
    _["sfstring_utf8_native"] = static_cast<int>(sfstring(native_utf8, 2, CE_NATIVE).encoding),
    _["sf_vec_utf8_native"] = static_cast<int>(vec_native_indexer.getCharLenCE(0).enc),
    _["slice_store_utf8_native"] = static_cast<int>(store_native_indexer.getCharLenCE(0).enc),
    _["direct_utf8_native"] = static_cast<int>(direct_indexer.getCharLenCE(0).enc)
  );
  UNPROTECT(4);
  return out;
}

// [[Rcpp::export]]
IntegerVector sf_external_get_encodings(SEXP x) {
  RStringIndexer indexer(x);
  const size_t len = indexer.size();
  IntegerVector out(len);
  for(size_t i = 0; i < len; ++i) {
    out[static_cast<R_xlen_t>(i)] = static_cast<int>(indexer.getCharLenCE(i).enc);
  }
  return out;
}

// [[Rcpp::export]]
List sf_external_make_storage_vectors() {
  const char utf8_e[] = "\\xC3\\xA9";
  const char latin1_e[] = "\\xE9";

  SEXP vec = PROTECT(sf_vector(6));
  sf_vec_data & vec_data = sf_vec_data_ref(vec);
  vec_data[0] = sfstring("abc", 3, cetype_t_ext::CE_ASCII);
  vec_data[1] = sfstring(utf8_e, 2, cetype_t_ext::CE_ASCII_OR_UTF8);
  vec_data[2] = sfstring(utf8_e, 2, cetype_t_ext::CE_NATIVE);
  vec_data[3] = sfstring(latin1_e, 1, cetype_t_ext::CE_LATIN1);
  vec_data[4] = sfstring("abc", 3, cetype_t_ext::CE_BYTES);
  vec_data[5] = sfstring(NA_STRING);

  SEXP store = PROTECT(slice_store_create_with_size(6, 256));
  slice_store_data & store_data = slice_store_data_ref(store);
  store_data.assign(0, "abc", 3, cetype_t_ext::CE_ASCII);
  store_data.assign(1, utf8_e, 2, cetype_t_ext::CE_ASCII_OR_UTF8);
  store_data.assign(2, utf8_e, 2, cetype_t_ext::CE_NATIVE);
  store_data.assign(3, latin1_e, 1, cetype_t_ext::CE_LATIN1);
  store_data.assign(4, "abc", 3, cetype_t_ext::CE_BYTES);
  store_data.assign(5, nullptr, 0, cetype_t_ext::CE_NA);

  List out = List::create(
    _["vec"] = vec,
    _["slice"] = store
  );
  UNPROTECT(2);
  return out;
}

// [[Rcpp::export]]
List sf_external_slice_store_state(SEXP x) {
  slice_store_data & data = slice_store_data_ref(x);
  return List::create(
    _["initial_slice_size"] = data.initial_slice_size_override.has_value() ?
      Rcpp::wrap(static_cast<int>(*data.initial_slice_size_override)) :
      Rcpp::wrap(NA_INTEGER),
    _["slice_count"] = static_cast<int>(data.slices.size()),
    _["current_slice_capacity"] = static_cast<int>(data.current_slice_capacity)
  );
}
'

is_solaris <- function() {
  grepl("SunOS", Sys.info()[["sysname"]])
}

can_source_cpp <- !is_solaris()

compile_regressions <- function() {
  if (!can_source_cpp) {
    return(invisible(NULL))
  }
  if (exists("sf_external_regressions", mode = "function", inherits = FALSE)) {
    return(invisible(NULL))
  }

  R_TESTS <- Sys.getenv("R_TESTS")
  if (nzchar(R_TESTS)) {
    R_TESTS_absolute <- normalizePath(R_TESTS)
    Sys.setenv(R_TESTS = R_TESTS_absolute)
  }
  Rcpp::sourceCpp(code = regression_external_source)
  if (nzchar(R_TESTS)) {
    Sys.setenv(R_TESTS = R_TESTS)
  }
  invisible(NULL)
}

encode_native_uint <- function(x, size) {
  remaining <- as.numeric(x)
  out <- as.raw(integer(size))
  for (i in seq_len(size)) {
    byte <- as.integer(remaining %% 256)
    out[[if (.Platform$endian == "little") i else size - i + 1L]] <- as.raw(byte)
    remaining <- floor(remaining / 256)
  }
  out
}

find_raw_subsequence <- function(haystack, needle) {
  stopifnot(is.raw(haystack), is.raw(needle), length(needle) > 0L)
  limit <- length(haystack) - length(needle) + 1L
  if (limit < 1L) {
    stop("raw subsequence longer than input")
  }
  for (i in seq_len(limit)) {
    if (identical(haystack[i:(i + length(needle) - 1L)], needle)) {
      return(i)
    }
  }
  stop("raw subsequence not found")
}

mutate_serialized_state <- function(x, expected_state, mutate) {
  serialized <- serialize(x, NULL, xdr = FALSE)
  start <- find_raw_subsequence(serialized, expected_state)
  state_idx <- start:(start + length(expected_state) - 1L)
  mutated_state <- mutate(serialized[state_idx])
  stopifnot(is.raw(mutated_state), length(mutated_state) == length(expected_state))
  serialized[state_idx] <- mutated_state
  serialized
}

run_regressions <- function(nthreads = 1L) {
  x <- convert_to_sf_vector(c("abcdef", "ghijkl"))
  y <- sf_substr(x, c(1L, 2L), 1L, nthreads = nthreads)
  stopifnot(string_identical(y, c("a", "")))
  materialize(x)
  y2 <- sf_substr(x, c(1L, 2L), 1L, nthreads = nthreads)
  stopifnot(string_identical(y2, c("a", "")))

  err <- tryCatch({
    stringfish:::sf_assign(sf_vector(2), 3, "x")
    NULL
  }, error = identity)
  stopifnot(inherits(err, "error"))

  collapsed <- sf_collapse(c("a", NA_character_), ",")
  stopifnot(is.character(collapsed), length(collapsed) == 1L, is.na(collapsed[[1]]))
  stopifnot(identical(sf_paste(nthreads = nthreads), ""))
  stopifnot(identical(sf_paste("a", "b", sep = NA_character_, nthreads = nthreads), NA_character_))
  stopifnot(identical(sf_collapse(c("a", "b"), NA_character_), NA_character_))

  stopifnot(string_identical(sf_tolower(c("ABC", NA_character_)), c("abc", NA_character_)))
  stopifnot(string_identical(sf_toupper(c("abc", NA_character_)), c("ABC", NA_character_)))

  stopifnot(identical(sf_compare("a", c("a", "b", NA_character_), nthreads = nthreads), c(TRUE, FALSE, NA)))
  stopifnot(identical(sf_compare(c("a", "b", NA_character_), "a", nthreads = nthreads), c(TRUE, FALSE, NA)))

  latin1_word <- iconv("café", "UTF-8", "latin1")
  Encoding(latin1_word) <- "latin1"
  bytes_word <- latin1_word
  Encoding(bytes_word) <- "bytes"
  normalized <- convert_to_sf_vector(latin1_word)
  stopifnot(identical(normalized, "café"), identical(Encoding(normalized), "UTF-8"))
  stopifnot(identical(sf_compare(c("café", bytes_word, "café"), latin1_word, nthreads = nthreads), c(TRUE, TRUE, TRUE)))

  tmp_latin1 <- tempfile(fileext = ".txt")
  con <- file(tmp_latin1, open = "wb")
  writeBin(c(charToRaw(latin1_word), as.raw(0x0a)), con)
  close(con)
  on.exit(unlink(tmp_latin1), add = TRUE)
  read_back <- sf_readLines(tmp_latin1, encoding = "latin1")
  stopifnot(identical(read_back, "café"), identical(Encoding(read_back), "UTF-8"))

  err <- tryCatch({
    sf_grepl("a", c("a", "b"), nthreads = nthreads)
    NULL
  }, error = identity)
  stopifnot(inherits(err, "error"))
  err <- tryCatch({
    sf_split("a", c("a", "b"), nthreads = nthreads)
    NULL
  }, error = identity)
  stopifnot(inherits(err, "error"))
  err <- tryCatch({
    sf_gsub("a", "a", c("x", "y"), nthreads = nthreads)
    NULL
  }, error = identity)
  stopifnot(inherits(err, "error"))

  subset_x <- convert_to_sf_vector(c("a", "b", "c"))
  stopifnot(string_identical(subset_x[c(3L, 1L, NA_integer_, 4L)], c("c", "a", NA_character_, NA_character_)))
  stopifnot(get_string_type(subset_x[c(3L, 1L)]) == "stringfish vector")
  materialize(subset_x)
  stopifnot(string_identical(subset_x[c(3L, 1L, NA_integer_, 4L)], c("c", "a", NA_character_, NA_character_)))
  stopifnot(get_string_type(subset_x[c(3L, 1L)]) == "stringfish vector")

  stopifnot(identical(
    materialize(sf_split("abc", "(?<=a)", nthreads = nthreads)[[1L]]),
    strsplit("abc", "(?<=a)", perl = TRUE)[[1L]]
  ))
  stopifnot(identical(
    materialize(sf_split("abc", "^", nthreads = nthreads)[[1L]]),
    c("", "abc")
  ))
  stopifnot(identical(
    materialize(sf_split("abc", "$", nthreads = nthreads)[[1L]]),
    c("abc", "")
  ))

  tmp_bytes <- tempfile(fileext = ".txt")
  on.exit(unlink(tmp_bytes), add = TRUE)
  bytes_word <- rawToChar(as.raw(0xE9))
  Encoding(bytes_word) <- "bytes"
  sf_writeLines(bytes_word, tmp_bytes, encode_mode = "UTF-8")
  stopifnot(identical(readBin(tmp_bytes, "raw", n = 2L), as.raw(c(0xE9, 0x0A))))

  if (can_source_cpp) {
    out <- sf_external_regressions(c("ab", NA_character_), nthreads)
    stopifnot(identical(out$nchar, c(2L, NA_integer_)))
    stopifnot(identical(out$compare, c(TRUE, NA)))
    stopifnot(string_identical(out$substr, c("a", NA_character_)))
    stopifnot(length(out$random) == 4L)

    native_word <- "é"
    Encoding(native_word) <- "unknown"
    stopifnot(identical(Encoding(native_word), "unknown"))
    enc <- sf_external_encoding_regressions(native_word)
    stopifnot(identical(unname(enc), c(3L, 3L, 3L, 3L, 0L, 0L, 0L, 0L, 0L)))

    sized_store <- stringfish:::slice_store_create_with_size(4L, 256L)
    sized_state <- sf_external_slice_store_state(sized_store)
    stopifnot(identical(sized_state$initial_slice_size, 256L))
    stopifnot(identical(sized_state$slice_count, 1L))
    stopifnot(identical(sized_state$current_slice_capacity, 256L))

    stored <- sf_external_make_storage_vectors()
    vec_before <- sf_external_get_encodings(stored$vec)
    slice_before <- sf_external_get_encodings(stored$slice)
    vec_after <- unserialize(serialize(stored$vec, NULL))
    slice_after <- unserialize(serialize(stored$slice, NULL))
    stopifnot(identical(sf_external_get_encodings(vec_after), vec_before))
    stopifnot(identical(sf_external_get_encodings(slice_after), slice_before))
    stopifnot(identical(get_string_type(vec_after), "stringfish vector"))
    stopifnot(identical(get_string_type(slice_after), "stringfish slice store"))

    serialized_state_abc <- c(
      encode_native_uint(1L, 8L),
      encode_native_uint(3L, 4L),
      as.raw(254L),
      charToRaw("abc")
    )
    truncated_serialized <- mutate_serialized_state(
      convert_to_sf_vector("abc"),
      serialized_state_abc,
      function(state) {
        state[[1L]] <- as.raw(0x02)
        state
      }
    )
    invalid_encoding_serialized <- mutate_serialized_state(
      convert_to_slice_store("abc"),
      serialized_state_abc,
      function(state) {
        state[[13L]] <- as.raw(0x04)
        state
      }
    )
    err <- tryCatch({
      unserialize(truncated_serialized)
      NULL
    }, error = identity)
    stopifnot(inherits(err, "error"))
    err <- tryCatch({
      unserialize(invalid_encoding_serialized)
      NULL
    }, error = identity)
    stopifnot(inherits(err, "error"))
    stopifnot(grepl("ALTREP", conditionMessage(err), fixed = TRUE))
  }
}

compile_regressions()
run_regressions(1L)
if (stringfish:::is_tbb()) {
  run_regressions(4L)
}

suppressPackageStartupMessages(library(stringfish))

alternate_case_source <- '
// [[Rcpp::depends(stringfish)]]
// [[Rcpp::plugins(cpp17)]]
#include <Rcpp.h>
#include "sf_external.h"
using namespace Rcpp;

// [[Rcpp::export]]
SEXP sf_alternate_case(SEXP x) {
  RStringIndexer r(x);
  size_t len = r.size();
  SEXP output = PROTECT(sf_vector(len));
  sf_vec_data & output_data = sf_vec_data_ref(output);
  size_t i = 0;
  for(auto e : r) {
    if(e.ptr == nullptr) {
      i++; // increment output index
      continue;
    }
    std::string temp(e.len, \'\\0\');
    bool case_switch = false;
    for(int j=0; j<e.len; j++) {
      if((e.ptr[j] >= 65) & (e.ptr[j] <= 90)) {
        if((case_switch = !case_switch)) {
          temp[j] = e.ptr[j] + 32;
          continue;
        }
      } else if((e.ptr[j] >= 97) & (e.ptr[j] <= 122)) {
        if(!(case_switch = !case_switch)) {
          temp[j] = e.ptr[j] - 32;
          continue;
        }
      } else if(e.ptr[j] == 32) {
        case_switch = false;
      }
      temp[j] = e.ptr[j];
    }
    output_data[i] = sfstring(temp, e.enc);
    i++;
  }
  UNPROTECT(1);
  return output;
}
'


is_solaris <- function() {
  grepl('SunOS',Sys.info()['sysname'])
}

can_source_cpp <- !is_solaris()

myfile <- tempfile()
print(myfile)

# For a test set, we are using the 500 most common Icelandic words
# This is a pretty good test set because all Icelandic words can be encoded as either UTF-8 or latin1. It also contains a mix of ASCII and non-ASCII strings
# https://en.wikipedia.org/wiki/ISO/IEC_8859-1
test_words_file <- system.file("icelandic_words_500_utf8.txt", package = "stringfish")
if (!nzchar(test_words_file)) {
  test_words_file <- file.path("inst", "icelandic_words_500_utf8.txt")
}
stopifnot(file.exists(test_words_file))
i500_utf8 <- readLines(test_words_file, encoding = "UTF-8", warn = FALSE)
i500_latin1 <- iconv(i500_utf8, "UTF-8", "latin1")

catn <- function(...) {
  cat(..., "\n")
}

split_list_identical <- function(x, y) {
  stopifnot(length(x) == length(y))
  all(vapply(seq_along(y), function(i) {
    string_identical(x[[i]], y[[i]])
  }, logical(1)))
}

to_latin1_list <- function(x) {
  lapply(x, function(z) iconv(z, from = "UTF-8", to = "latin1"))
}

ntests <- 50

use_tbb <- stringfish:::is_tbb()
cat("using thread building blocks?", use_tbb, "\n")
if(use_tbb) {
  nthreads <- c(1,4)
} else {
  nthreads <- 1
}

print(sessionInfo())
print(utils::localeToCharset())

if(can_source_cpp) {
  R_TESTS <- Sys.getenv("R_TESTS") # startup.Rs
  if (nzchar(R_TESTS)) {
    R_TESTS_absolute <- normalizePath(R_TESTS)
    Sys.setenv(R_TESTS = R_TESTS_absolute)
  }
  Rcpp::sourceCpp(code = alternate_case_source)
  if (nzchar(R_TESTS)) Sys.setenv(R_TESTS = R_TESTS)
} else {
  catn("skipping sourceCpp-backed tests")
}
for(.j in 1:3) {
  cat("iteration", .j, "\n")
  if(.j %% 2 == 0) {
    stringfish:::set_is_utf8_locale()
  } else {
    stringfish:::unset_is_utf8_locale()
  }
  for(nt in nthreads) {
    cat("number of threads", nt, "\n")

    catn("test altrep serialization (base::serialize)")
    x <- convert_to_sf_vector(sample(c(i500_utf8, i500_latin1), size = length(i500_utf8) + length(i500_latin1), replace=TRUE))
    y <- unserialize(serialize(x, NULL))
    stopifnot(string_identical(x,y))
    stopifnot(get_string_type(y) == "stringfish vector")
    x_slice <- convert_to_slice_store(sample(c(i500_utf8, i500_latin1), size = length(i500_utf8) + length(i500_latin1), replace=TRUE))
    y_slice <- unserialize(serialize(x_slice, NULL))
    stopifnot(string_identical(x_slice, y_slice))
    stopifnot(get_string_type(y_slice) == "stringfish slice store")

    catn("sf_assign")
    for(. in 1:ntests) {
      # if(. == 1) {gctorture(TRUE)} else {gctorture(FALSE)}
      x <- sf_vector(10)
      y <- character(10)
      for(i in 1:10) {
        new_str <- sample(c(i500_utf8,i500_latin1),1)
        if(.j %% 2 == 1) materialize(x)
        stringfish:::sf_assign(x, i, new_str)
        y[i] <- new_str
      }
      stopifnot(string_identical(x,y))
    }
      
      catn("sf_iconv")
      for(. in 1:ntests) {
        # if(. == 1) {gctorture(TRUE)} else {gctorture(FALSE)}
        x <- sf_iconv(i500_latin1, "latin1", "UTF-8")
        if(.j %% 2 == 1) materialize(x)
        y <- sf_iconv(i500_utf8, "UTF-8", "latin1")
        stopifnot(string_identical(x, i500_utf8))
        stopifnot(string_identical(y, i500_latin1))
        y <- sf_iconv(convert_to_sf_vector(i500_utf8), "UTF-8", "latin1")
        x <- sf_iconv(y, "latin1", "UTF-8")
        stopifnot(string_identical(x, i500_utf8))
        stopifnot(string_identical(y, i500_latin1))
      }
      
      catn("sf_nchar")
      for(. in 1:ntests) {
        # if(. == 1) {gctorture(TRUE)} else {gctorture(FALSE)}
        x <- sf_iconv(i500_utf8, "UTF-8", "latin1")
        if(.j %% 2 == 1) materialize(x)
        y <- convert_to_sf_vector(i500_utf8)
        stopifnot( identical(sf_nchar(x, nthreads = nt), nchar(x)) )
        stopifnot( identical(sf_nchar(y, nthreads = nt), nchar(y)) )
        stopifnot( identical(sf_nchar(x, nthreads = nt), nchar(i500_latin1)) )
        stopifnot( identical(sf_nchar(y, nthreads = nt), nchar(i500_utf8)) )
        stopifnot( identical(sf_nchar(x, type = "bytes", nthreads = nt), nchar(i500_latin1, type = "bytes")) )
        stopifnot( identical(sf_nchar(y, type = "bytes", nthreads = nt), nchar(i500_utf8, type = "bytes")) )
      }
      
      catn("sf_substr")
      for(. in 1:ntests) {
        # if(. == 1) {gctorture(TRUE)} else {gctorture(FALSE)}
        start <- sample(-10:10, size=1)
        if(start < 0) {
          rstart <- sf_nchar(i500_latin1, nthreads = nt) + start + 1
        } else {
          rstart <- start
        }
        stop <- sample(-10:10, size=1)
        if(stop < 0) {
          rstop <- sf_nchar(i500_latin1, nthreads = nt) + stop + 1
        } else {
          rstop <- stop
        }
        x <- sf_substr(i500_latin1, start, stop, nthreads = nt)
        if(.j %% 2 == 1) materialize(x)
        y <- substr(i500_latin1, rstart, rstop)
        x2 <- sf_substr(i500_utf8, start, stop, nthreads = nt)
        y2 <- substr(i500_utf8, rstart, rstop)
        stopifnot(string_identical(x, y))
        stopifnot(string_identical(x2, y2))
      }
      
      catn("sf_collapse")
      for(. in 1:ntests) {
        # if(. == 1) {gctorture(TRUE)} else {gctorture(FALSE)}
        x <- sf_collapse(i500_latin1, collapse = ":::")
        if(.j %% 2 == 1) materialize(x)
        y <- paste0(i500_latin1, collapse = ":::")
        # stopifnot(string_identical(x, y)) # paste0 converts to UTF-8 -- doesn't respect encoding
        stopifnot(x == y)
        x <- sf_collapse(i500_latin1, collapse = ",")
        if(.j %% 2 == 1) materialize(x)
        y <- paste0(i500_latin1, collapse = ",")
        stopifnot(x == y)
        x <- sf_collapse(i500_utf8, collapse = ":::")
        if(.j %% 2 == 1) materialize(x)
        y <- paste0(i500_utf8, collapse = ":::")
        stopifnot(x == y)
        x <- sf_collapse(i500_utf8, collapse = ",")
        if(.j %% 2 == 1) materialize(x)
        y <- paste0(i500_utf8, collapse = ",")
        stopifnot(x == y)
      }
      
      catn("sf_paste")
      for(. in 1:ntests) {
        # if(. == 1) {gctorture(TRUE)} else {gctorture(FALSE)}
        x <- do.call(paste, c(as.list(i500_latin1), sep=":::"))
        y <- do.call(sf_paste, c(as.list(i500_latin1), sep=":::", nthreads = nt))
        if(.j %% 2 == 1) materialize(y)
        stopifnot(x == y)
        x <- do.call(paste, c(as.list(i500_latin1), sep=":::"))
        y <- do.call(sf_paste, c(as.list(i500_latin1), sep=":::", nthreads = nt))
        if(.j %% 2 == 1) materialize(y)
        stopifnot(x == y)
        x <- do.call(paste, c(as.list(i500_utf8), sep=","))
        y <- do.call(sf_paste, c(as.list(i500_utf8), sep=",", nthreads = nt))
        if(.j %% 2 == 1) materialize(y)
        stopifnot(x == y)
        x <- do.call(paste, c(as.list(i500_utf8), sep=","))
        y <- do.call(sf_paste, c(as.list(i500_utf8), sep=",", nthreads = nt))
        if(.j %% 2 == 1) materialize(y)
        stopifnot(x == y)
      }
      
      catn("sf_readLines")
      for(. in 1:ntests) {
        # if(. == 1) {gctorture(TRUE)} else {gctorture(FALSE)}
        writeLines(i500_utf8, con=myfile, useBytes=T)
        x <- sf_readLines(myfile, encoding = "UTF-8")
        if(.j %% 2 == 1) materialize(x)
        y <- readLines(myfile); Encoding(y) <- "UTF-8"
        stopifnot(string_identical(x, y))
        writeLines(i500_latin1, con=myfile)
        x <- sf_readLines(myfile, encoding = "latin1")
        if(.j %% 2 == 1) materialize(x)
        y <- readLines(myfile); Encoding(y) <- "latin1"
        stopifnot(string_identical(x, y))
      }

      catn("sf_writeLines")
      for(. in 1:ntests) {
        write_utf8 <- tempfile()
        on.exit(unlink(write_utf8), add = TRUE)
        sf_writeLines(i500_utf8, write_utf8, encode_mode = "UTF-8")
        y <- readLines(write_utf8, warn = FALSE, encoding = "UTF-8")
        Encoding(y) <- "UTF-8"
        stopifnot(string_identical(y, i500_utf8))

        write_bytes <- tempfile()
        on.exit(unlink(write_bytes), add = TRUE)
        bytes_word <- rawToChar(as.raw(0xE9))
        Encoding(bytes_word) <- "bytes"
        sf_writeLines(bytes_word, write_bytes, encode_mode = "UTF-8")
        stopifnot(identical(readBin(write_bytes, "raw", n = 2L), as.raw(c(0xE9, 0x0A))))
      }
      
      catn("sf_grepl")
      for(. in 1:ntests) {
        # if(. == 1) {gctorture(TRUE)} else {gctorture(FALSE)}
        p <- rawToChar(as.raw(c(0x5e, 0xc3, 0xb6, 0x2e, 0x2b)))
        Encoding(p) <- "UTF-8"
        p2 <- rawToChar(as.raw(c(0x5e, 0xf6, 0x2e, 0x2b)))
        Encoding(p2) <- "latin1"
        stopifnot(all(sf_grepl(i500_utf8, p, nthreads = nt) == grepl(p, i500_utf8)))
        stopifnot(all(sf_grepl(i500_latin1, p2, nthreads = nt) == grepl(p2, i500_latin1)))
        
        stopifnot(sf_grepl(i500_utf8, "[a-f]", nthreads = nt) == grepl("[a-f]", i500_utf8))
        stopifnot(sf_grepl(i500_latin1, "[a-f]", nthreads = nt) == grepl("[a-f]", i500_latin1))
      }
      
      catn("sf_grepl fixed")
      for(. in 1:ntests) {
        # if(. == 1) {gctorture(TRUE)} else {gctorture(FALSE)}
        p <- rawToChar(as.raw(c(0xc3, 0xb6)))
        Encoding(p) <- "UTF-8"
        p2 <- rawToChar(as.raw(c(0xf6)))
        Encoding(p2) <- "latin1"
        stopifnot(all(sf_grepl(i500_utf8, p, fixed = T, nthreads = nt) == grepl(p, i500_utf8)))
        stopifnot(all(sf_grepl(i500_latin1, p2, fixed = T,  nthreads = nt) == grepl(p2, i500_latin1)))
        
        stopifnot(sf_grepl(i500_utf8, "[a-f]", nthreads = nt) == grepl("[a-f]", i500_utf8))
        stopifnot(sf_grepl(i500_latin1, "[a-f]", nthreads = nt) == grepl("[a-f]", i500_latin1))
      }
      
      catn("sf_gsub")
      for(. in 1:ntests) {
        # if(. == 1) {gctorture(TRUE)} else {gctorture(FALSE)}
        p <- rawToChar(as.raw(c(0x5e, 0xc3, 0xb6, 0x2e, 0x2b, 0x28, 0x2e, 0x29, 0x24)))
        Encoding(p) <- "UTF-8"
        p2 <- rawToChar(as.raw(c(0x5e, 0xf6, 0x2e, 0x2b, 0x28, 0x2e, 0x29, 0x24)))
        Encoding(p2) <- "latin1"
        stopifnot(all(sf_gsub(i500_utf8, p, "$1", nthreads = nt) == gsub(p, "\\1", i500_utf8)))
        stopifnot(all(sf_gsub(i500_latin1, p2, "$1", nthreads = nt) == gsub(p2, "\\1", i500_latin1)))
        
        p <- "^h.+(.)$"
        stopifnot(all(sf_gsub(i500_utf8, p, "$1", nthreads = nt) == gsub(p, "\\1", i500_utf8)))
        stopifnot(all(sf_gsub(i500_latin1, p, "$1", nthreads = nt) == gsub(p, "\\1", i500_latin1)))
      }
      
      catn("sf_split")
      for(. in 1:ntests) {
        x <- sf_split(i500_utf8, "", nthreads = nt)
        stopifnot(split_list_identical(x, strsplit(i500_utf8, "", perl = TRUE)))

        x <- sf_split(rep("", 1e3), "a", nthreads=nt)
        stopifnot(all(x == ""))

        x <- sf_split(rep("", 1e3), "", nthreads=nt)
        stopifnot(all(x == ""))

        x <- sf_split(rep("abcde", 1e3), "f", nthreads=nt)
        stopifnot(all(x == "abcde"))

        utf8_subject <- c("café", "éclair", "plain", "é", "été")
        utf8_expected <- list(c("caf", ""), c("", "clair"), "plain", c("", ""), c("", "t", ""))
        x <- sf_split(utf8_subject, "é", nthreads = nt)
        stopifnot(split_list_identical(x, utf8_expected))

        utf8_regex_subject <- c("héj", "méow", "plain", "éx")
        utf8_regex_expected <- list(c("h", ""), c("m", "w"), "plain", c("", ""))
        x <- sf_split(utf8_regex_subject, "é.", nthreads = nt)
        stopifnot(split_list_identical(x, utf8_regex_expected))

        latin1_subject <- iconv(utf8_regex_subject, from = "UTF-8", to = "latin1")
        latin1_expected <- to_latin1_list(utf8_regex_expected)
        latin1_pattern <- iconv("é.", from = "UTF-8", to = "latin1")
        x <- sf_split(latin1_subject, latin1_pattern, nthreads = nt)
        x <- lapply(x, sf_iconv, from = "UTF-8", to = "latin1")
        stopifnot(split_list_identical(x, latin1_expected))

        x <- sf_split(latin1_subject, latin1_pattern, encode_mode = "byte", nthreads = nt)
        x <- lapply(x, function(z) {
          z <- materialize(z)
          Encoding(z) <- "latin1"
          z
        })
        stopifnot(split_list_identical(x, latin1_expected))

        zero_width <- materialize(sf_split("abc", "(?<=a)", nthreads = nt)[[1L]])
        stopifnot(identical(zero_width, strsplit("abc", "(?<=a)", perl = TRUE)[[1L]]))

        zero_width_start <- materialize(sf_split("abc", "^", nthreads = nt)[[1L]])
        stopifnot(identical(zero_width_start, c("", "abc")))

        zero_width_end <- materialize(sf_split("abc", "$", nthreads = nt)[[1L]])
        stopifnot(identical(zero_width_end, c("abc", "")))
      }
      
      catn("sf_toupper and sf_tolower")
      for(. in 1:ntests) {
        # if(. == 1) {gctorture(TRUE)} else {gctorture(FALSE)}
        x1 <- sf_toupper(i500_latin1)
        x2 <- sf_toupper(i500_utf8)
        y1 <- sf_tolower(i500_latin1)
        if(.j %% 2 == 1) materialize(y1)
        y2 <- sf_tolower(i500_utf8)
        if(.j %% 2 == 1) materialize(y2)
        z1 <- sf_tolower(x1)
        z2 <- sf_tolower(x2)
        stopifnot(string_identical(z1, i500_latin1))
        stopifnot(string_identical(z2, i500_utf8))
        stopifnot(string_identical(y1, i500_latin1))
        stopifnot(string_identical(y2, i500_utf8))
        # base R functions also convert Unicode characters to upper
        # stopifnot(string_identical(x1, iconv(toupper(i500_latin1),"UTF-8", "latin1")))
        # stopifnot(string_identical(x2, toupper(i500_utf8)))
      }
      
      if(can_source_cpp) {
        catn("Rcpp test with sf_alternate_case")
        for(. in 1:ntests) {
          x <- c("hello world", "HELLO WORLD")
          string_identical(sf_alternate_case(x), c("hElLo wOrLd", "hElLo wOrLd"))
        }
        
        catn("sf_trim")
        for(. in 1:ntests) {
          # if(. == 1) {gctorture(TRUE)} else {gctorture(FALSE)}
          x <- sf_trim(sf_paste("\t", i500_utf8, " \n"))
          if(.j %% 2 == 1) materialize(x)
          stopifnot(string_identical(x, i500_utf8))
          
          x <- sf_trim(sf_iconv(sf_paste("\t", i500_utf8, " \n"), from = "UTF-8", to = "latin1"), encode_mode = "byte")
          x <- materialize(x)
          Encoding(x) <- "latin1"
          stopifnot(string_identical(x, i500_latin1))
        }
      }
      
      # Disable check due to https://bugs.r-project.org/show_bug.cgi?id=18211
      catn("sf_match")
      # gctorture(TRUE)
      for(. in 1:ntests) {
        i500_utf8_shuffled <- c(NA_character_, i500_utf8[sample(length(i500_utf8))][-1])
        temp <- c(i500_utf8, NA_character_)
        x <- sf_match(temp, i500_utf8_shuffled)
        # y <- match(temp, i500_utf8_shuffled)
        # stopifnot(identical(x,y))
        i500_latin1_shuffled <- c(NA_character_, i500_latin1[sample(length(i500_latin1))][-1])
        temp <- c(i500_latin1, NA_character_)
        x <- sf_match(c(i500_latin1, NA_character_), i500_latin1_shuffled)
        # y <- match(temp, i500_latin1_shuffled)
        # stopifnot(identical(x,y))
      }
      # gctorture(FALSE)
      
      catn("sf_compare")
      for(. in 1:ntests) {
        # if(. == 1) {gctorture(TRUE)} else {gctorture(FALSE)}
        i500_utf8_shuffled <- i500_utf8
        i500_utf8_shuffled[sample(length(i500_utf8), size = 100)] <- ""
        x <- sf_compare(c(i500_utf8, NA_character_), c(i500_utf8_shuffled, NA_character_))
        y <- c(i500_utf8, NA_character_) == c(i500_utf8_shuffled, NA_character_)
        stopifnot(identical(x,y))
        
        i500_latin1_shuffled <- i500_latin1
        i500_latin1_shuffled[sample(length(i500_latin1), size = 100)] <- ""
        x <- sf_compare(c(i500_latin1, NA_character_), c(i500_latin1_shuffled, NA_character_))
        y <- c(i500_latin1, NA_character_) == c(i500_latin1_shuffled, NA_character_)
        stopifnot(identical(x,y))
      }

      catn("sf_concat")
      for(. in 1:ntests) {
        # if(. == 1) {gctorture(TRUE)} else {gctorture(FALSE)}
        i500_utf8_shuffled <- i500_utf8
        i500_utf8_shuffled[sample(length(i500_utf8), size = 100)] <- ""
        x <- sfc(sfc(i500_utf8, NA_character_), sfc(i500_utf8_shuffled, NA_character_), character(0))
        if(.j %% 2 == 1) materialize(x)
        y <- sfc(sfc(convert_to_sf_vector(i500_utf8), NA_character_), sfc(convert_to_sf_vector(i500_utf8_shuffled), NA_character_), character(0))
        z <- c(c(convert_to_sf_vector(i500_utf8), NA_character_), c(convert_to_sf_vector(i500_utf8_shuffled), NA_character_), character(0))
        z2 <- c(c(i500_utf8, NA_character_), c(i500_utf8_shuffled, NA_character_), character(0))
        stopifnot(string_identical(x,y))
        stopifnot(string_identical(x,z))
        stopifnot(string_identical(x,z2))
        stopifnot(string_identical(x,y))
        stopifnot(identical(sfc(character(0)), character(0)))
      }
      print(gc())
    }
  }
print("end")

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

set_is_utf8_locale <- function() {
    invisible(.Call(`_stringfish_set_is_utf8_locale`))
}

unset_is_utf8_locale <- function() {
    invisible(.Call(`_stringfish_unset_is_utf8_locale`))
}

get_is_utf8_locale <- function() {
    .Call(`_stringfish_get_is_utf8_locale`)
}

is_tbb <- function() {
    .Call(`_stringfish_is_tbb`)
}

check_simd <- function() {
    invisible(.Call(`_stringfish_check_simd`))
}

get_pcre2_info <- function() {
    .Call(`_stringfish_get_pcre2_info`)
}

get_string_type <- function(x) {
    .Call(`_stringfish_get_string_type`, x)
}

materialize <- function(x) {
    .Call(`_stringfish_materialize`, x)
}

sf_vector <- function(len) {
    .Call(`_stringfish_sf_vector`, len)
}

sf_vector_create <- function(len) {
    .Call(`_stringfish_sf_vector_create`, len)
}

slice_store_create <- function(len) {
    .Call(`_stringfish_slice_store_create`, len)
}

slice_store_create_with_size <- function(len, initial_slice_size) {
    .Call(`_stringfish_slice_store_create_with_size`, len, initial_slice_size)
}

sf_assign <- function(x, i, e) {
    invisible(.Call(`_stringfish_sf_assign`, x, i, e))
}

sf_iconv <- function(x, from, to, nthreads = getOption("stringfish.nthreads", 1L)) {
    .Call(`_stringfish_sf_iconv`, x, from, to, nthreads)
}

convert_to_sf_vector <- function(x, length_out = length(x)) {
    .Call(`_stringfish_convert_to_sf_vector`, x, length_out)
}

convert_to_sf <- function(x) {
    .Call(`_stringfish_convert_to_sf`, x)
}

convert_to_slice_store <- function(x, length_out = length(x)) {
    .Call(`_stringfish_convert_to_slice_store`, x, length_out)
}

sf_nchar <- function(x, type = "chars", nthreads = getOption("stringfish.nthreads", 1L)) {
    .Call(`_stringfish_sf_nchar`, x, type, nthreads)
}

sf_substr <- function(x, start, stop, nthreads = getOption("stringfish.nthreads", 1L)) {
    .Call(`_stringfish_sf_substr`, x, start, stop, nthreads)
}

c_sf_paste <- function(dots, sep, nthreads = 1L) {
    .Call(`_stringfish_c_sf_paste`, dots, sep, nthreads)
}

sf_collapse <- function(x, collapse) {
    .Call(`_stringfish_sf_collapse`, x, collapse)
}

sf_readLines <- function(file, encoding = "UTF-8") {
    .Call(`_stringfish_sf_readLines`, file, encoding)
}

sf_writeLines <- function(text, file, sep = "\n", na_value = "NA", encode_mode = "UTF-8") {
    invisible(.Call(`_stringfish_sf_writeLines`, text, file, sep, na_value, encode_mode))
}

sf_grepl <- function(subject, pattern, encode_mode = "auto", fixed = FALSE, nthreads = getOption("stringfish.nthreads", 1L)) {
    .Call(`_stringfish_sf_grepl`, subject, pattern, encode_mode, fixed, nthreads)
}

sf_split <- function(subject, split, encode_mode = "auto", fixed = FALSE, nthreads = getOption("stringfish.nthreads", 1L)) {
    .Call(`_stringfish_sf_split`, subject, split, encode_mode, fixed, nthreads)
}

sf_gsub <- function(subject, pattern, replacement, encode_mode = "auto", fixed = FALSE, nthreads = getOption("stringfish.nthreads", 1L)) {
    .Call(`_stringfish_sf_gsub`, subject, pattern, replacement, encode_mode, fixed, nthreads)
}

random_strings <- function(N, string_size = 50L, charset = "abcdefghijklmnopqrstuvwxyz", vector_mode = "stringfish") {
    .Call(`_stringfish_random_strings`, N, string_size, charset, vector_mode)
}

sf_tolower <- function(x) {
    .Call(`_stringfish_sf_tolower`, x)
}

sf_toupper <- function(x) {
    .Call(`_stringfish_sf_toupper`, x)
}

sf_match <- function(x, table, nthreads = getOption("stringfish.nthreads", 1L)) {
    .Call(`_stringfish_sf_match`, x, table, nthreads)
}

sf_compare <- function(x, y, nthreads = getOption("stringfish.nthreads", 1L)) {
    .Call(`_stringfish_sf_compare`, x, y, nthreads)
}

c_sf_concat <- function(x) {
    .Call(`_stringfish_c_sf_concat`, x)
}

# stringfish - Quick Serialization of R Objects
# Copyright (C) 2020-present Travers Ching
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# 
# You can contact the author at:
#   https://github.com/traversc/stringfish


#' sf_vector
#' 
#' Creates a new empty stringfish vector
#' @usage sf_vector(len)
#' @param len length of the new vector
#' @return A new empty stringfish vector
#' @details 
#' This is a backwards-compatible alias for `sf_vector_create(len)`.
#' It creates a new stringfish vector, an alt-rep character vector backed by a C++ "std::vector" as the internal memory representation.
#' The vector type is "sfstring", which is a simple C++ class containing owned string bytes and a single byte (uint8_t) representing the encoding. 
#' @examples
#' x <- sf_vector(10)
#' sf_assign(x, 1, "hello world")
#' sf_assign(x, 2, "another string")
#' @name sf_vector
NULL

#' sf_vector_create
#' 
#' Creates a new empty stringfish vector
#' @usage sf_vector_create(len)
#' @param len length of the new vector
#' @return A new stringfish vector
#' @details 
#' This function creates a new empty `sf_vec`-backed stringfish vector.
#' If you want to fill the vector from character data, use `convert_to_sf_vector`.
#' @examples
#' x <- sf_vector_create(4)
#' @name sf_vector_create
NULL

#' slice_store_create
#'
#' Creates a new empty slice-store-backed stringfish vector
#' @usage slice_store_create(len)
#' @param len length of the new vector
#' @return A new slice-store-backed stringfish vector
#' @details
#' This function creates a new stringfish vector backed by `slice_store`,
#' which stores string bytes in append-only slices plus per-element records.
#' If you want to fill the vector from character data, use `convert_to_slice_store`.
#' @examples
#' x <- slice_store_create(4)
#' @name slice_store_create
NULL

#' slice_store_create_with_size
#'
#' Creates a new empty slice-store-backed stringfish vector with a fixed initial slice size
#' @usage slice_store_create_with_size(len, initial_slice_size)
#' @param len length of the new vector
#' @param initial_slice_size Initial size of the first underlying `slice_store` slice.
#' @return A new slice-store-backed stringfish vector
#' @details
#' This function creates a new stringfish vector backed by `slice_store`,
#' and uses `initial_slice_size` for the first slice allocation instead of the default heuristic.
#' If you want to fill the vector from character data, use `convert_to_slice_store`.
#' @examples
#' x <- slice_store_create_with_size(4, 256)
#' @name slice_store_create_with_size
NULL

#' sf_assign
#' 
#' Assigns a new string to a stringfish vector or any other character vector
#' @usage sf_assign(x, i, e)
#' @param x the vector
#' @param i the index to assign to
#' @param e the new string to replace at i in x
#' @return No return value, the function assigns an element to an existing stringfish vector
#' @details 
#' A function to assign a new element to an existing character vector. If the the vector is a stringfish vector, it does so without materialization. 
#' @examples
#' x <- sf_vector(10)
#' sf_assign(x, 1, "hello world")
#' sf_assign(x, 2, "another string")
#' @name sf_assign
NULL

#' get_string_type
#' 
#' Returns the type of the character vector
#' @usage get_string_type(x)
#' @param x the vector
#' @return The type of vector
#' @details 
#' A function that returns the type of character vector. Possible values are "normal vector", "stringfish vector",
#' "stringfish vector (materialized)", "stringfish slice store", "stringfish slice store (materialized)"
#' or "other alt-rep vector"
#' @examples
#' x <- sf_vector(10)
#' get_string_type(x) # returns "stringfish vector"
#' x <- character(10)
#' get_string_type(x) # returns "normal vector"
#' @name get_string_type
NULL

#' materialize
#' 
#' Materializes an alt-rep object
#' @usage materialize(x)
#' @param x An alt-rep object
#' @return x
#' @details 
#' Materializes any alt-rep object and then returns it. 
#' Note: the object is materialized regardless of whether the return value is assigned to a variable. 
#' @examples
#' x <- sf_vector(10)
#' sf_assign(x, 1, "hello world")
#' sf_assign(x, 2, "another string")
#' x <- materialize(x)
#' @name materialize
NULL

#' convert_to_sf_vector
#' 
#' Converts a character vector to an `sf_vec`-backed stringfish vector
#' `convert_to_sf()` is a compatibility alias for `convert_to_sf_vector()`.
#' @usage convert_to_sf_vector(x, length.out = length(x))
#' @aliases convert_to_sf
#' @param x A character vector
#' @param length.out Optional output length used to recycle `x`
#' @return The converted character vector
#' @details 
#' Converts a character vector to a stringfish vector backed by `sf_vec`.
#' If `length.out` is supplied, `x` is recycled to that length before conversion.
#' The opposite of `materialize`.
#' @examples
#' x <- convert_to_sf_vector(letters)
#' @name convert_to_sf_vector
NULL

#' convert_to_slice_store
#' 
#' Converts a character vector to a slice-store-backed stringfish vector
#' @usage convert_to_slice_store(x, length.out = length(x))
#' @param x A character vector
#' @param length.out Optional output length used to recycle `x`
#' @return The converted character vector
#' @details
#' Converts a character vector to a stringfish vector backed by `slice_store`.
#' If `length.out` is supplied, `x` is recycled to that length before conversion.
#' The converter pre-sizes the first `slice_store` slice from the normalized string bytes when possible.
#' The opposite of `materialize`.
#' @examples
#' x <- convert_to_slice_store(letters)
#' @name convert_to_slice_store
NULL

#' sf_iconv
#' 
#' Converts encoding of one character vector to another
#' @usage sf_iconv(x, from, to, nthreads = getOption("stringfish.nthreads", 1L))
#' @param x An alt-rep object
#' @param from the encoding to assume of `x`
#' @param nthreads Number of threads to use
#' @param to the new encoding
#' @return the converted character vector as a stringfish vector
#' @details 
#' This is an analogue to the base R function `iconv`. It converts a string from one encoding (e.g. latin1 or UTF-8) to another
#' @seealso iconv
#' @examples
#' x <- "fa\xE7ile"
#' Encoding(x) <- "latin1"
#' sf_iconv(x, "latin1", "UTF-8")
#' @name sf_iconv
NULL

#' sf_nchar
#' 
#' Counts the number of characters in a character vector
#' @usage sf_nchar(x, type = "chars", nthreads = getOption("stringfish.nthreads", 1L))
#' @param x A character vector
#' @param type The type of counting to perform ("chars" or "bytes", default: "chars")
#' @param nthreads Number of threads to use
#' @return An integer vector of the number of characters
#' @details 
#' Returns the number of characters per string. The type of counting only matters for UTF-8 strings, where a character can be represented by multiple bytes. 
#' @seealso nchar
#' @examples
#' x <- "fa\xE7ile"
#' Encoding(x) <- "latin1"
#' x <- sf_iconv(x, "latin1", "UTF-8")
#' @name sf_nchar
NULL

#' sf_substr
#' 
#' Extracts substrings from a character vector
#' @usage sf_substr(x, start, stop, nthreads = getOption("stringfish.nthreads", 1L))
#' @param x A character vector
#' @param start The begining to extract from
#' @param stop The end to extract from
#' @param nthreads Number of threads to use
#' @return A stringfish vector of substrings
#' @details 
#' This works the same way as `substr`, but in addition allows negative indexing. 
#' Negative indicies count backwards from the end of the string, with -1 being the last character. 
#' @seealso substr
#' @examples
#' x <- c("fa\xE7ile", "hello world")
#' Encoding(x) <- "latin1"
#' x <- sf_iconv(x, "latin1", "UTF-8")
#' sf_substr(x, 4, -1) # extracts from the 4th character to the last
#' ## [1] "ile"  "lo world"
#' @name sf_substr
NULL

#' sf_collapse
#' 
#' Pastes a series of strings together separated by the `collapse` parameter
#' @usage sf_collapse(x, collapse)
#' @param x A character vector
#' @param collapse A single string
#' @return A single string with all values in `x` pasted together, separated by `collapse`.
#' @details 
#' This works the same way as `paste0(x, collapse=collapse)`
#' @seealso paste0, paste
#' @examples
#' x <- c("hello", "\\xe4\\xb8\\x96\\xe7\\x95\\x8c")
#' Encoding(x) <- "UTF-8"
#' sf_collapse(x, " ") # "hello world" in Japanese
#' sf_collapse(letters, "") # returns the alphabet
#' @name sf_collapse
NULL

#' sf_paste
#' 
#' Pastes a series of strings together
#' @usage sf_paste(..., sep = "", nthreads = getOption("stringfish.nthreads", 1L))
#' @param ... Any number of character vector strings
#' @param sep The seperating string between strings
#' @param nthreads Number of threads to use
#' @return A character vector where elements of the arguments are pasted together
#' @details 
#' This works the same way as `paste0(..., sep=sep)`
#' @seealso paste0, paste
#' @examples
#' x <- letters
#' y <- LETTERS
#' sf_paste(x,y, sep = ":")
#' @name sf_paste
NULL

#' sf_readLines
#' 
#' A function that reads a file line by line
#' @usage sf_readLines(file, encoding = "UTF-8")
#' @param file The file name
#' @param encoding The encoding to use (Default: UTF-8)
#' @return A stringfish vector of the lines in a file
#' @details 
#' A function for reading in text data using `std::ifstream`.
#' @seealso readLines
#' @examples
#' file <- tempfile()
#' sf_writeLines(letters, file)
#' sf_readLines(file)
#' @name sf_readLines
NULL

#' sf_writeLines
#' 
#' A function that writes text line by line
#' @usage sf_writeLines(text, file, sep = "\n", na_value = "NA", encode_mode = "UTF-8")
#' @param text A character to write to file
#' @param file Name of the file to write to
#' @param sep The line separator character(s)
#' @param na_value What to write in case of a NA string
#' @param encode_mode "UTF-8" or "byte". If "UTF-8", text strings are normalized to UTF-8 while `CE_BYTES` strings are written as raw bytes.
#' @details 
#' A function for writing text data using `std::ofstream`.
#' @seealso writeLines
#' @examples
#' file <- tempfile()
#' sf_writeLines(letters, file)
#' sf_readLines(file)
#' @name sf_writeLines
NULL

#' sf_grepl
#' 
#' A function that matches patterns and returns a logical vector
#' @usage sf_grepl(subject, pattern, encode_mode = "auto", fixed = FALSE, 
#' nthreads = getOption("stringfish.nthreads", 1L))
#' @param subject The subject character vector to search
#' @param pattern The pattern to search for
#' @param encode_mode "auto", "UTF-8" or "byte". Determines multi-byte (UTF-8) characters or single-byte characters are used.
#' @param fixed determines whether the pattern parameter should be interpreted literally or as a regular expression
#' @param nthreads Number of threads to use
#' @return A logical vector with the same length as subject
#' @details 
#' The function uses the PCRE2 library, which is also used internally by R. 
#' The encoding is based on the pattern string (or forced via the encode_mode parameter). 
#' Note: the order of paramters is switched compared to the `grepl` base R function, with subject being first. 
#' @seealso grepl
#' @examples
#' x <- sf_vector(10)
#' sf_assign(x, 1, "hello world")
#' pattern <- "^hello"
#' sf_grepl(x, pattern)
#' @name sf_grepl
NULL

#' sf_gsub
#' 
#' A function that performs pattern substitution
#' @usage sf_gsub(subject, pattern, replacement, encode_mode = "auto", fixed = FALSE, 
#' nthreads = getOption("stringfish.nthreads", 1L))
#' @param subject The subject character vector to search
#' @param pattern The pattern to search for
#' @param replacement The replacement string
#' @param encode_mode "auto", "UTF-8" or "byte". Determines multi-byte (UTF-8) characters or single-byte characters are used.
#' @param fixed determines whether the pattern parameter should be interpreted literally or as a regular expression
#' @param nthreads Number of threads to use
#' @return A stringfish vector of the replacement string
#' @details 
#' The function uses the PCRE2 library, which is also used internally by R. However, syntax may be slightly different. 
#' E.g.: capture groups: "\1" in R, but "$1" in PCRE2 (as in Perl). 
#' The encoding of the output is determined by the pattern (or forced using encode_mode parameter) and encodings should be compatible. 
#' E.g: mixing ASCII and UTF-8 is okay, but not UTF-8 and latin1. 
#' Note: the order of paramters is switched compared to the `gsub` base R function, with subject being first. 
#' @seealso gsub
#' @examples
#' x <- "hello world"
#' pattern <- "^hello (.+)"
#' replacement <- "goodbye $1"
#' sf_gsub(x, pattern, replacement)
#' @name sf_gsub
NULL

#' random_strings
#' 
#' A function that generates random strings
#' @usage random_strings(N, string_size = 50L, charset = "abcdefghijklmnopqrstuvwxyz",
#'                       vector_mode = "stringfish")
#' @param N The number of strings to generate
#' @param string_size Either a single non-negative integer applied to every output
#' string, or a non-negative integer vector of length `N`.
#' @param charset The characters used to generate the random strings (default: abcdefghijklmnopqrstuvwxyz)
#' @param vector_mode The type of character vector to generate (either stringfish or normal, default: stringfish)
#' @return A character vector of the random strings
#' @details 
#' A convenience function for generating test strings.
#' @examples
#' set.seed(1)
#' x <- random_strings(1e6, 80L, "ACGT", vector_mode = "stringfish")
#' y <- random_strings(4, c(1L, 2L, 4L, 8L), "ACGT")
#' @name random_strings
NULL

#' sf_toupper
#' 
#' A function converting a string to all uppercase
#' @usage sf_toupper(x)
#' @param x A character vector
#' @return A stringfish vector where all lowercase is converted to uppercase
#' @details 
#' Note: the function only converts ASCII characters. 
#' @seealso toupper
#' @examples
#' x <- letters
#' sf_toupper(x)
#' @name sf_toupper
NULL

#' sf_tolower
#' 
#' A function converting a string to all lowercase
#' @usage sf_tolower(x)
#' @param x A character vector
#' @return A stringfish vector where all uppercase is converted to lowercase
#' @details 
#' Note: the function only converts ASCII characters. 
#' @seealso tolower
#' @examples
#' x <- LETTERS
#' sf_tolower(x)
#' @name sf_tolower
NULL

#' sf_starts
#' 
#' A function for detecting a pattern at the start of a string
#' @usage sf_starts(subject, pattern, ...)
#' @param subject A character vector
#' @param pattern A string to look for at the start
#' @param ... Parameters passed to sf_grepl
#' @return A logical vector true if there is a match, false if no match, NA is the subject was NA
#' @seealso startsWith, sf_ends
#' @examples
#' x <- c("alpha", "beta", "gamma", "delta", "epsilon")
#' sf_starts(x, "a")
#' @name sf_starts
NULL

#' sf_ends
#' 
#' A function for detecting a pattern at the end of a string
#' @usage sf_ends(subject, pattern, ...)
#' @param subject A character vector
#' @param pattern A string to look for at the start
#' @param ... Parameters passed to sf_grepl
#' @return A logical vector true if there is a match, false if no match, NA is the subject was NA
#' @seealso endsWith, sf_starts
#' @examples
#' x <- c("alpha", "beta", "gamma", "delta", "epsilon")
#' sf_ends(x, "a")
#' @name sf_ends
NULL

#' sf_trim
#' 
#' A function to remove leading/trailing whitespace
#' @usage sf_trim(subject, which = c("both", "left", "right"), whitespace = "[ \\\\t\\\\r\\\\n]", ...)
#' @param subject A character vector
#' @param which "both", "left", or "right" determines which white space is removed
#' @param whitespace Whitespace characters (default: "[ \\\\t\\\\r\\\\n]")
#' @param ... Parameters passed to sf_gsub
#' @return A stringfish vector of trimmed whitespace
#' @seealso trimws
#' @examples
#' x <- c(" alpha ", " beta", " gamma ", "delta ", "epsilon ")
#' sf_trim(x)
#' @name sf_trim
NULL

#' sf_match
#' 
#' Returns a vector of the positions of x in table
#' @usage sf_match(x, table, nthreads = getOption("stringfish.nthreads", 1L))
#' @param x A character vector to search for in table
#' @param table A character vector to be matched against x
#' @param nthreads Number of threads to use
#' @return An integer vector of the indicies of each x element's position in table
#' @seealso match
#' @details Note: similarly to the base R function, long "table" vectors are not supported. This is due to the maximum integer value that can be returned (`.Machine$integer.max`)
#' @examples
#' sf_match("c", letters)
#' @name sf_match
NULL

#' sf_split
#' 
#' A function to split strings by a delimiter
#' @usage sf_split(subject, split, encode_mode = "auto", fixed = FALSE, 
#' nthreads = getOption("stringfish.nthreads", 1L))
#' @param subject A character vector
#' @param split A delimiter to split the string by
#' @param encode_mode "auto", "UTF-8" or "byte". Determines multi-byte (UTF-8) characters or single-byte characters are used.
#' @param fixed determines whether the split parameter should be interpreted literally or as a regular expression
#' @param nthreads Number of threads to use
#' @return A list of stringfish character vectors
#' @seealso strsplit
#' @examples
#' sf_split(datasets::state.name, "\\s") # split U.S. state names by any space character
#' @name sf_split
NULL

#' sf_compare
#' 
#' Returns a logical vector testing equality of strings from two string vectors
#' @usage sf_compare(x, y, nthreads = getOption("stringfish.nthreads", 1L))
#' @param x A character vector of length 1 or the same non-zero length as y
#' @param y Another character vector of length 1 or the same non-zero length as y
#' @param nthreads Number of threads to use
#' @return A logical vector
#' @details Note: the function tests semantic string equality. Non-byte text is normalized to a UTF-8 working representation, while `CE_BYTES` strings are compared byte-for-byte.
#' @examples
#' sf_compare(letters, "a")
#' @name sf_compare
NULL

#' @rdname sf_compare
sf_equals <- sf_compare

#' string_identical
#' 
#' Compare strings semantically or exactly
#' @usage string_identical(x, y, mode = c("semantic", "exact"))
#' @param x A character vector
#' @param y Another character vector to compare to x
#' @param mode Either `"semantic"` to compare text after normalizing non-byte
#' strings to UTF-8, or `"exact"` to additionally require matching encoding.
#' Strings marked as `"bytes"` are always compared exactly.
#' @return TRUE if strings are identical under the selected comparison mode
#' @seealso identical
#' @examples
#' x <- "fa\xE7ile"
#' Encoding(x) <- "latin1"
#' y <- iconv(x, "latin1", "UTF-8")
#' identical(x, y) # TRUE
#' string_identical(x, y) # TRUE
#' string_identical(x, y, mode = "exact") # FALSE
#' @name string_identical
NULL

#' sf_concat
#' 
#' Appends vectors together
#' @usage sf_concat(...)
#' @param ... Any number of vectors, coerced to character vector if necessary
#' @return A concatenated stringfish vector
#' @examples
#' sf_concat(letters, 1:5)
#' @name sf_concat
NULL

#' @rdname sf_concat
sfc <- sf_concat

sf_vector <- function(len) {
  sf_vector_create(len)
}

sf_vector_create <- function(len) {
  .Call(`_stringfish_sf_vector_create`, len)
}

slice_store_create <- function(len) {
  .Call(`_stringfish_slice_store_create`, len)
}

slice_store_create_with_size <- function(len, initial_slice_size) {
  if(!is.numeric(initial_slice_size) ||
     length(initial_slice_size) != 1L ||
     is.na(initial_slice_size) ||
     initial_slice_size < 0 ||
     initial_slice_size != floor(initial_slice_size)) {
    stop("initial_slice_size must be a non-negative whole number")
  }
  .Call(`_stringfish_slice_store_create_with_size`, len, initial_slice_size)
}

convert_to_sf_vector <- function(x, length.out = length(x)) {
  if(!is.character(x)) {
    stop("x must be a character vector")
  }
  x_len <- length(x)
  if(length(length.out) != 1L ||
     is.na(length.out) ||
     length.out < 0 ||
     length.out != floor(length.out)) {
    stop("length.out must be a non-negative whole number")
  }
  if(length.out > 0L && x_len < 1L) {
    stop("x must have length >= 1 when length.out is > 0")
  }
  .Call(`_stringfish_convert_to_sf_vector`, x, length.out)
}

convert_to_slice_store <- function(x, length.out = length(x)) {
  if(!is.character(x)) {
    stop("x must be a character vector")
  }
  x_len <- length(x)
  if(length(length.out) != 1L ||
     is.na(length.out) ||
     length.out < 0 ||
     length.out != floor(length.out)) {
    stop("length.out must be a non-negative whole number")
  }
  if(length.out > 0L && x_len < 1L) {
    stop("x must have length >= 1 when length.out is > 0")
  }
  .Call(`_stringfish_convert_to_slice_store`, x, length.out)
}

sf_paste <- function(..., sep="", nthreads = getOption("stringfish.nthreads", 1L)) {
  if(!is.character(sep) || length(sep) != 1) {
    stop("sep should be a character vector of length 1")
  }
  dots <- list(...)
  len <- -1
  for(i in seq_along(dots)) {
    if(!is.character(dots[[i]])) {
      dots[[i]] <- as.character(dots[[i]])
    }
    li <- length(dots[[i]])
    if(li == 0) stop("argument cannot be of length zero")
    if(li == 1) next
    if(len == -1) {
      len <- li
    } else {
      if(li != len) stop("All arguments should be the same length or length 1")
    }
  }
  c_sf_paste(dots, sep, nthreads)
}

sf_concat <- function(...) {
  dots <- list(...)
  for(i in seq_along(dots)) {
    if(!is.character(dots[[i]])) dots[[i]] <- as.character(dots[[i]])
  }
  c_sf_concat(dots)
}

sf_starts <- function(subject, pattern, ...) {
  pattern <- paste0("^", pattern)
  sf_grepl(subject, pattern, ...)
}

sf_ends <- function(subject, pattern, ...) {
  pattern <- paste0(pattern, "$")
  sf_grepl(subject, pattern, ...)
}

sf_trim <- function(subject, which = c("both", "left", "right"), whitespace = "[ \\t\\r\\n]", ...) {
  which <- match.arg(which)
  if(which == "both") {
    sf_gsub(sf_gsub(subject, paste0("^", whitespace,"+"), "", ...), paste0(whitespace, "+", "$"), "", ...)
  } else if(which == "left") {
    sf_gsub(subject, paste0("^", whitespace, "+"), "", ...)
  } else {
    sf_gsub(subject, paste0(whitespace, "+", "$"), "", ...)
  }
}

string_identical <- function(x, y, mode = c("semantic", "exact")) {
  stopifnot(is.character(x))
  stopifnot(is.character(y))
  mode <- match.arg(mode)
  if(length(x) != length(y)) return(FALSE)
  na_x <- is.na(x)
  na_y <- is.na(y)
  if(!identical(na_x, na_y)) return(FALSE)
  if(all(na_x)) return(TRUE) # correctly catches zero length as well
  x <- x[!na_x]
  y <- y[!na_y]

  if(mode == "exact") {
    if(any(nchar(x) != nchar(y))) return(FALSE)
    if(!all(Encoding(x) == Encoding(y))) return(FALSE)
    if(any(x != y)) return(FALSE)
    return(TRUE)
  }

  x_bytes <- Encoding(x) == "bytes"
  y_bytes <- Encoding(y) == "bytes"
  if(any(x_bytes | y_bytes)) {
    if(!identical(x_bytes, y_bytes)) return(FALSE)
    if(any(x_bytes)) {
      ok_bytes <- mapply(identical, as.list(x[x_bytes]), as.list(y[y_bytes]), USE.NAMES = FALSE)
      if(!all(ok_bytes)) return(FALSE)
    }
  }
  if(all(x_bytes)) return(TRUE)
  identical(unname(enc2utf8(x[!x_bytes])), unname(enc2utf8(y[!y_bytes])))
}

---
title: "stringfish"
output:
  html_vignette:
    keep_md: no
  rmarkdown::github_document: default
vignette: >
  %\VignetteIndexEntry{stringfish}
  \usepackage[utf8]{inputenc}
  %\VignetteEngine{knitr::rmarkdown}
---

```{r, setup, echo=FALSE}
IS_GITHUB <- Sys.getenv("IS_GITHUB") != ""
```

```{r results='asis', echo=FALSE, eval=IS_GITHUB}
cat('
[![R-CMD-check](https://github.com/traversc/stringfish/workflows/R-CMD-check/badge.svg)](https://github.com/traversc/stringfish/actions)
[![CRAN-Status-Badge](https://www.r-pkg.org/badges/version/stringfish)](https://cran.r-project.org/package=stringfish)
[![CRAN-Downloads-Badge](https://cranlogs.r-pkg.org/badges/stringfish)](https://cran.r-project.org/package=stringfish)
[![CRAN-Downloads-Total-Badge](https://cranlogs.r-pkg.org/badges/grand-total/stringfish)](https://cran.r-project.org/package=stringfish)
')
```

`stringfish` is a framework for string and sequence operations using the ALTREP system (introduced in R 3.5) as a way to represent R objects using custom memory layout.

This package has two primary goals:

- Provide R users a way to speed up common string operations compared to base R (see benchmarks below)
- Create a common ALTREP framework that can be used by other packages with full interoperability

`stringfish` currently provides two ALTREP backends with the same semantics: `sf_vec`, a simple vector of string objects, and `slice_store`, which stores strings within large contiguous blocks of memory. They make different storage tradeoffs, but the same `stringfish` operations work across both.

For text data, `stringfish` is intentionally UTF-8-centric outside of explicit byte mode, so conversions, comparisons, and ALTREP views stay consistent across normal R vectors and both backends.

## Installation
```{r eval=FALSE}
install.packages("stringfish", type="source", configure.args="--with-simd=AVX2")
```

## Benchmark

The simplest way to show the utility of the ALTREP framework is through a quick benchmark comparing `stringfish` and base R. 

```{r echo=FALSE, results='asis'}
if(IS_GITHUB) {
  cat('![](vignettes/bench_v3.png "bench_v3"){width=576px}')
} else {
  cat('![](bench_v3.png "bench_v3"){width=576px}')
}
```

On favorable workloads, some functions in `stringfish` can be more than an order of magnitude faster than vectorized base R operations, and built-in multithreading can widen that gap further. On large text datasets, this can turn minutes of computation into seconds. 

## Currently implemented functions

A list of implemented `stringfish` functions and analogous base R functions:

* `sf_iconv` (`iconv`)
* `sf_nchar` (`nchar`)
* `sf_substr` (`substr`)
* `sf_paste` (`paste0`)
* `sf_collapse` (`paste0`)
* `sf_readLines` (`readLines`)
* `sf_writeLines` (`writeLines`)
* `sf_grepl` (`grepl`)
* `sf_gsub` (`gsub`)
* `sf_toupper` (`toupper`)
* `sf_tolower` (`tolower`)
* `sf_starts` (`startsWith`)
* `sf_ends` (`endsWith`)
* `sf_trim` (`trimws`)
* `sf_split` (`strsplit`)
* `sf_match` (`match` for strings only)
* `sf_compare`/`sf_equals` (`==`, ALTREP-aware semantic string equality)
* `sf_concat`/`sfc` (`c`)

Utility functions:

* `sf_vector_create` -- creates a new empty `sf_vec`-backed stringfish vector
* `sf_vector` -- backwards-compatible alias for `sf_vector_create`
* `slice_store_create` -- creates a new empty `slice_store`-backed stringfish vector
* `slice_store_create_with_size` -- creates a `slice_store`-backed stringfish vector with an explicit initial slice size
* `sf_assign` -- assign strings into a `stringfish` vector in place (like `x[i] <- "mystring"`)
* `convert_to_sf_vector` -- converts a character vector to a `stringfish` vector
* `convert_to_slice_store` -- converts a character vector to a `stringfish` slice store
* `get_string_type` -- determines string type (whether ALTREP or normal)
* `materialize` -- converts any ALTREP object into a normal R object
* `random_strings` -- creates random strings as either a `stringfish` or normal R vector
* `string_identical` -- compares strings either semantically or exactly across encodings

In addition, many R operations in base R and other packages are already ALTREP-aware (i.e. they don't cause materialization). Functions that subset or index into string vectors generally do not materialize.

* `sample`
* `head`
* `tail`
* `[` -- e.g. `x[20:30]`
* various tidyverse filters and operations
* Etc.

`stringfish` functions are not intended to exactly replicate their base R analogues. One difference is that `subject` parameters are always the first argument, which is easier to use with pipes. E.g., `gsub(pattern, replacement, subject)` becomes `sf_gsub(subject, pattern, replacement)`.

## Extensibility

`stringfish` as a framework is intended to be easily extensible. Stringfish vectors can be worked into `Rcpp` scripts or even into other packages. The example below creates an `sf_vec`-backed output because it is simple and direct, but the same indexing semantics work across both backends.

Below is a detailed `Rcpp` script that creates a function to alternate upper and lower case of strings. 

```{c eval=FALSE}
// [[Rcpp::depends(stringfish)]]
#include <Rcpp.h>
#include "sf_external.h"
using namespace Rcpp;

// [[Rcpp::export]]
SEXP sf_alternate_case(SEXP x) {
  // Iterate through a character vector using the RStringIndexer class
  // If the input vector x is a stringfish character vector it will do so without materialization
  RStringIndexer r(x);
  size_t len = r.size();
  
  // Create an output stringfish vector
  // Like all R objects, it must be protected from garbage collection
  SEXP output = PROTECT(sf_vector_create(len));
  
  // Obtain a reference to the underlying output data
  sf_vec_data & output_data = sf_vec_data_ref(output);
  
  // You can use range based for loop via an iterator class that returns RStringIndexer::rstring_info e
  // rstring info is a struct containing const char * ptr, int len, and an encoding flag
  // ptr should be treated as a byte pointer plus length, not as a null-terminated C string
  // a NA string is represented by a nullptr
  // Alternatively, access the data via the function r.getCharLenCE(i)
  size_t i = 0;
  for(auto e : r) {
    // check if string is NA and go to next if it is
    if(e.ptr == nullptr) {
      i++; // increment output index
      continue;
    }
    // Create a temporary output string and process the results.
    // This example intentionally toggles ASCII letters only.
    std::string temp(e.len, '\0');
    bool case_switch = false;
    for(int j=0; j<e.len; j++) {
      if((e.ptr[j] >= 65) && (e.ptr[j] <= 90)) { // char j is upper case
        if((case_switch = !case_switch)) { // check if we should convert to lower case
          temp[j] = e.ptr[j] + 32;
          continue;
        }
      } else if((e.ptr[j] >= 97) && (e.ptr[j] <= 122)) { // char j is lower case
        if(!(case_switch = !case_switch)) { // check if we should convert to upper case
          temp[j] = e.ptr[j] - 32;
          continue;
        }
      } else if(e.ptr[j] == 32) {
        case_switch = false;
      }
      temp[j] = e.ptr[j];
    }
    
    // Create a new vector element sfstring and insert the processed string into the stringfish vector
    // sfstring has three constructors, 1) taking a std::string and encoding, 
    // 2) a char pointer and encoding, or 3) a CHARSXP object (e.g. sfstring(NA_STRING))
    output_data[i] = sfstring(temp, e.enc);
    i++; // increment output index
  }
  // Finally, call unprotect and return result
  UNPROTECT(1);
  return output;
}

```

Example function call:
```{r eval=FALSE}
sf_alternate_case("hello world") 
[1] "hElLo wOrLd"
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zz_help_files.R
\name{sf_paste}
\alias{sf_paste}
\title{sf_paste}
\usage{
sf_paste(..., sep = "", nthreads = getOption("stringfish.nthreads", 1L))
}
\arguments{
\item{...}{Any number of character vector strings}

\item{sep}{The seperating string between strings}

\item{nthreads}{Number of threads to use}
}
\value{
A character vector where elements of the arguments are pasted together
}
\description{
Pastes a series of strings together
}
\details{
This works the same way as `paste0(..., sep=sep)`
}
\examples{
x <- letters
y <- LETTERS
sf_paste(x,y, sep = ":")
}
\seealso{
paste0, paste
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zz_help_files.R
\name{slice_store_create_with_size}
\alias{slice_store_create_with_size}
\title{slice_store_create_with_size}
\usage{
slice_store_create_with_size(len, initial_slice_size)
}
\arguments{
\item{len}{length of the new vector}

\item{initial_slice_size}{Initial size of the first underlying `slice_store` slice.}
}
\value{
A new slice-store-backed stringfish vector
}
\description{
Creates a new empty slice-store-backed stringfish vector with a fixed initial slice size
}
\details{
This function creates a new stringfish vector backed by `slice_store`,
and uses `initial_slice_size` for the first slice allocation instead of the default heuristic.
If you want to fill the vector from character data, use `convert_to_slice_store`.
}
\examples{
x <- slice_store_create_with_size(4, 256)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zz_help_files.R
\name{sf_vector}
\alias{sf_vector}
\title{sf_vector}
\usage{
sf_vector(len)
}
\arguments{
\item{len}{length of the new vector}
}
\value{
A new empty stringfish vector
}
\description{
Creates a new empty stringfish vector
}
\details{
This is a backwards-compatible alias for `sf_vector_create(len)`.
It creates a new stringfish vector, an alt-rep character vector backed by a C++ "std::vector" as the internal memory representation.
The vector type is "sfstring", which is a simple C++ class containing owned string bytes and a single byte (uint8_t) representing the encoding.
}
\examples{
x <- sf_vector(10)
sf_assign(x, 1, "hello world")
sf_assign(x, 2, "another string")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zz_help_files.R
\name{sf_compare}
\alias{sf_compare}
\alias{sf_equals}
\title{sf_compare}
\usage{
sf_compare(x, y, nthreads = getOption("stringfish.nthreads", 1L))

sf_equals(x, y, nthreads = getOption("stringfish.nthreads", 1L))
}
\arguments{
\item{x}{A character vector of length 1 or the same non-zero length as y}

\item{y}{Another character vector of length 1 or the same non-zero length as y}

\item{nthreads}{Number of threads to use}
}
\value{
A logical vector
}
\description{
Returns a logical vector testing equality of strings from two string vectors
}
\details{
Note: the function tests semantic string equality. Non-byte text is normalized to a UTF-8 working representation, while `CE_BYTES` strings are compared byte-for-byte.
}
\examples{
sf_compare(letters, "a")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zz_help_files.R
\name{sf_gsub}
\alias{sf_gsub}
\title{sf_gsub}
\usage{
sf_gsub(subject, pattern, replacement, encode_mode = "auto", fixed = FALSE, 
nthreads = getOption("stringfish.nthreads", 1L))
}
\arguments{
\item{subject}{The subject character vector to search}

\item{pattern}{The pattern to search for}

\item{replacement}{The replacement string}

\item{encode_mode}{"auto", "UTF-8" or "byte". Determines multi-byte (UTF-8) characters or single-byte characters are used.}

\item{fixed}{determines whether the pattern parameter should be interpreted literally or as a regular expression}

\item{nthreads}{Number of threads to use}
}
\value{
A stringfish vector of the replacement string
}
\description{
A function that performs pattern substitution
}
\details{
The function uses the PCRE2 library, which is also used internally by R. However, syntax may be slightly different. 
E.g.: capture groups: "\1" in R, but "$1" in PCRE2 (as in Perl). 
The encoding of the output is determined by the pattern (or forced using encode_mode parameter) and encodings should be compatible. 
E.g: mixing ASCII and UTF-8 is okay, but not UTF-8 and latin1. 
Note: the order of paramters is switched compared to the `gsub` base R function, with subject being first.
}
\examples{
x <- "hello world"
pattern <- "^hello (.+)"
replacement <- "goodbye $1"
sf_gsub(x, pattern, replacement)
}
\seealso{
gsub
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zz_help_files.R
\name{get_string_type}
\alias{get_string_type}
\title{get_string_type}
\usage{
get_string_type(x)
}
\arguments{
\item{x}{the vector}
}
\value{
The type of vector
}
\description{
Returns the type of the character vector
}
\details{
A function that returns the type of character vector. Possible values are "normal vector", "stringfish vector",
"stringfish vector (materialized)", "stringfish slice store", "stringfish slice store (materialized)"
or "other alt-rep vector"
}
\examples{
x <- sf_vector(10)
get_string_type(x) # returns "stringfish vector"
x <- character(10)
get_string_type(x) # returns "normal vector"
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zz_help_files.R
\name{sf_assign}
\alias{sf_assign}
\title{sf_assign}
\usage{
sf_assign(x, i, e)
}
\arguments{
\item{x}{the vector}

\item{i}{the index to assign to}

\item{e}{the new string to replace at i in x}
}
\value{
No return value, the function assigns an element to an existing stringfish vector
}
\description{
Assigns a new string to a stringfish vector or any other character vector
}
\details{
A function to assign a new element to an existing character vector. If the the vector is a stringfish vector, it does so without materialization.
}
\examples{
x <- sf_vector(10)
sf_assign(x, 1, "hello world")
sf_assign(x, 2, "another string")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zz_help_files.R
\name{slice_store_create}
\alias{slice_store_create}
\title{slice_store_create}
\usage{
slice_store_create(len)
}
\arguments{
\item{len}{length of the new vector}
}
\value{
A new slice-store-backed stringfish vector
}
\description{
Creates a new empty slice-store-backed stringfish vector
}
\details{
This function creates a new stringfish vector backed by `slice_store`,
which stores string bytes in append-only slices plus per-element records.
If you want to fill the vector from character data, use `convert_to_slice_store`.
}
\examples{
x <- slice_store_create(4)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zz_help_files.R
\name{sf_vector_create}
\alias{sf_vector_create}
\title{sf_vector_create}
\usage{
sf_vector_create(len)
}
\arguments{
\item{len}{length of the new vector}
}
\value{
A new stringfish vector
}
\description{
Creates a new empty stringfish vector
}
\details{
This function creates a new empty `sf_vec`-backed stringfish vector.
If you want to fill the vector from character data, use `convert_to_sf_vector`.
}
\examples{
x <- sf_vector_create(4)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zz_help_files.R
\name{sf_concat}
\alias{sf_concat}
\alias{sfc}
\title{sf_concat}
\usage{
sf_concat(...)

sfc(...)
}
\arguments{
\item{...}{Any number of vectors, coerced to character vector if necessary}
}
\value{
A concatenated stringfish vector
}
\description{
Appends vectors together
}
\examples{
sf_concat(letters, 1:5)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zz_help_files.R
\name{sf_nchar}
\alias{sf_nchar}
\title{sf_nchar}
\usage{
sf_nchar(x, type = "chars", nthreads = getOption("stringfish.nthreads", 1L))
}
\arguments{
\item{x}{A character vector}

\item{type}{The type of counting to perform ("chars" or "bytes", default: "chars")}

\item{nthreads}{Number of threads to use}
}
\value{
An integer vector of the number of characters
}
\description{
Counts the number of characters in a character vector
}
\details{
Returns the number of characters per string. The type of counting only matters for UTF-8 strings, where a character can be represented by multiple bytes.
}
\examples{
x <- "fa\xE7ile"
Encoding(x) <- "latin1"
x <- sf_iconv(x, "latin1", "UTF-8")
}
\seealso{
nchar
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zz_help_files.R
\name{materialize}
\alias{materialize}
\title{materialize}
\usage{
materialize(x)
}
\arguments{
\item{x}{An alt-rep object}
}
\value{
x
}
\description{
Materializes an alt-rep object
}
\details{
Materializes any alt-rep object and then returns it. 
Note: the object is materialized regardless of whether the return value is assigned to a variable.
}
\examples{
x <- sf_vector(10)
sf_assign(x, 1, "hello world")
sf_assign(x, 2, "another string")
x <- materialize(x)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zz_help_files.R
\name{sf_tolower}
\alias{sf_tolower}
\title{sf_tolower}
\usage{
sf_tolower(x)
}
\arguments{
\item{x}{A character vector}
}
\value{
A stringfish vector where all uppercase is converted to lowercase
}
\description{
A function converting a string to all lowercase
}
\details{
Note: the function only converts ASCII characters.
}
\examples{
x <- LETTERS
sf_tolower(x)
}
\seealso{
tolower
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zz_help_files.R
\name{sf_split}
\alias{sf_split}
\title{sf_split}
\usage{
sf_split(subject, split, encode_mode = "auto", fixed = FALSE, 
nthreads = getOption("stringfish.nthreads", 1L))
}
\arguments{
\item{subject}{A character vector}

\item{split}{A delimiter to split the string by}

\item{encode_mode}{"auto", "UTF-8" or "byte". Determines multi-byte (UTF-8) characters or single-byte characters are used.}

\item{fixed}{determines whether the split parameter should be interpreted literally or as a regular expression}

\item{nthreads}{Number of threads to use}
}
\value{
A list of stringfish character vectors
}
\description{
A function to split strings by a delimiter
}
\examples{
sf_split(datasets::state.name, "\\\\s") # split U.S. state names by any space character
}
\seealso{
strsplit
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zz_help_files.R
\name{sf_iconv}
\alias{sf_iconv}
\title{sf_iconv}
\usage{
sf_iconv(x, from, to, nthreads = getOption("stringfish.nthreads", 1L))
}
\arguments{
\item{x}{An alt-rep object}

\item{from}{the encoding to assume of `x`}

\item{nthreads}{Number of threads to use}

\item{to}{the new encoding}
}
\value{
the converted character vector as a stringfish vector
}
\description{
Converts encoding of one character vector to another
}
\details{
This is an analogue to the base R function `iconv`. It converts a string from one encoding (e.g. latin1 or UTF-8) to another
}
\examples{
x <- "fa\xE7ile"
Encoding(x) <- "latin1"
sf_iconv(x, "latin1", "UTF-8")
}
\seealso{
iconv
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zz_help_files.R
\name{sf_collapse}
\alias{sf_collapse}
\title{sf_collapse}
\usage{
sf_collapse(x, collapse)
}
\arguments{
\item{x}{A character vector}

\item{collapse}{A single string}
}
\value{
A single string with all values in `x` pasted together, separated by `collapse`.
}
\description{
Pastes a series of strings together separated by the `collapse` parameter
}
\details{
This works the same way as `paste0(x, collapse=collapse)`
}
\examples{
x <- c("hello", "\\\\xe4\\\\xb8\\\\x96\\\\xe7\\\\x95\\\\x8c")
Encoding(x) <- "UTF-8"
sf_collapse(x, " ") # "hello world" in Japanese
sf_collapse(letters, "") # returns the alphabet
}
\seealso{
paste0, paste
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zz_help_files.R
\name{sf_grepl}
\alias{sf_grepl}
\title{sf_grepl}
\usage{
sf_grepl(subject, pattern, encode_mode = "auto", fixed = FALSE, 
nthreads = getOption("stringfish.nthreads", 1L))
}
\arguments{
\item{subject}{The subject character vector to search}

\item{pattern}{The pattern to search for}

\item{encode_mode}{"auto", "UTF-8" or "byte". Determines multi-byte (UTF-8) characters or single-byte characters are used.}

\item{fixed}{determines whether the pattern parameter should be interpreted literally or as a regular expression}

\item{nthreads}{Number of threads to use}
}
\value{
A logical vector with the same length as subject
}
\description{
A function that matches patterns and returns a logical vector
}
\details{
The function uses the PCRE2 library, which is also used internally by R. 
The encoding is based on the pattern string (or forced via the encode_mode parameter). 
Note: the order of paramters is switched compared to the `grepl` base R function, with subject being first.
}
\examples{
x <- sf_vector(10)
sf_assign(x, 1, "hello world")
pattern <- "^hello"
sf_grepl(x, pattern)
}
\seealso{
grepl
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zz_help_files.R
\name{sf_trim}
\alias{sf_trim}
\title{sf_trim}
\usage{
sf_trim(subject, which = c("both", "left", "right"), whitespace = "[ \\\\t\\\\r\\\\n]", ...)
}
\arguments{
\item{subject}{A character vector}

\item{which}{"both", "left", or "right" determines which white space is removed}

\item{whitespace}{Whitespace characters (default: "[ \\\\t\\\\r\\\\n]")}

\item{...}{Parameters passed to sf_gsub}
}
\value{
A stringfish vector of trimmed whitespace
}
\description{
A function to remove leading/trailing whitespace
}
\examples{
x <- c(" alpha ", " beta", " gamma ", "delta ", "epsilon ")
sf_trim(x)
}
\seealso{
trimws
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zz_help_files.R
\name{random_strings}
\alias{random_strings}
\title{random_strings}
\usage{
random_strings(N, string_size = 50L, charset = "abcdefghijklmnopqrstuvwxyz",
                      vector_mode = "stringfish")
}
\arguments{
\item{N}{The number of strings to generate}

\item{string_size}{Either a single non-negative integer applied to every output
string, or a non-negative integer vector of length `N`.}

\item{charset}{The characters used to generate the random strings (default: abcdefghijklmnopqrstuvwxyz)}

\item{vector_mode}{The type of character vector to generate (either stringfish or normal, default: stringfish)}
}
\value{
A character vector of the random strings
}
\description{
A function that generates random strings
}
\details{
A convenience function for generating test strings.
}
\examples{
set.seed(1)
x <- random_strings(1e6, 80L, "ACGT", vector_mode = "stringfish")
y <- random_strings(4, c(1L, 2L, 4L, 8L), "ACGT")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zz_help_files.R
\name{sf_match}
\alias{sf_match}
\title{sf_match}
\usage{
sf_match(x, table, nthreads = getOption("stringfish.nthreads", 1L))
}
\arguments{
\item{x}{A character vector to search for in table}

\item{table}{A character vector to be matched against x}

\item{nthreads}{Number of threads to use}
}
\value{
An integer vector of the indicies of each x element's position in table
}
\description{
Returns a vector of the positions of x in table
}
\details{
Note: similarly to the base R function, long "table" vectors are not supported. This is due to the maximum integer value that can be returned (`.Machine$integer.max`)
}
\examples{
sf_match("c", letters)
}
\seealso{
match
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zz_help_files.R
\name{sf_readLines}
\alias{sf_readLines}
\title{sf_readLines}
\usage{
sf_readLines(file, encoding = "UTF-8")
}
\arguments{
\item{file}{The file name}

\item{encoding}{The encoding to use (Default: UTF-8)}
}
\value{
A stringfish vector of the lines in a file
}
\description{
A function that reads a file line by line
}
\details{
A function for reading in text data using `std::ifstream`.
}
\examples{
file <- tempfile()
sf_writeLines(letters, file)
sf_readLines(file)
}
\seealso{
readLines
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zz_help_files.R
\name{convert_to_sf_vector}
\alias{convert_to_sf_vector}
\alias{convert_to_sf}
\title{convert_to_sf_vector}
\usage{
convert_to_sf_vector(x, length.out = length(x))
}
\arguments{
\item{x}{A character vector}

\item{length.out}{Optional output length used to recycle `x`}
}
\value{
The converted character vector
}
\description{
Converts a character vector to an `sf_vec`-backed stringfish vector
`convert_to_sf()` is a compatibility alias for `convert_to_sf_vector()`.
}
\details{
Converts a character vector to a stringfish vector backed by `sf_vec`.
If `length.out` is supplied, `x` is recycled to that length before conversion.
The opposite of `materialize`.
}
\examples{
x <- convert_to_sf_vector(letters)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zz_help_files.R
\name{sf_substr}
\alias{sf_substr}
\title{sf_substr}
\usage{
sf_substr(x, start, stop, nthreads = getOption("stringfish.nthreads", 1L))
}
\arguments{
\item{x}{A character vector}

\item{start}{The begining to extract from}

\item{stop}{The end to extract from}

\item{nthreads}{Number of threads to use}
}
\value{
A stringfish vector of substrings
}
\description{
Extracts substrings from a character vector
}
\details{
This works the same way as `substr`, but in addition allows negative indexing. 
Negative indicies count backwards from the end of the string, with -1 being the last character.
}
\examples{
x <- c("fa\xE7ile", "hello world")
Encoding(x) <- "latin1"
x <- sf_iconv(x, "latin1", "UTF-8")
sf_substr(x, 4, -1) # extracts from the 4th character to the last
## [1] "ile"  "lo world"
}
\seealso{
substr
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zz_help_files.R
\name{sf_writeLines}
\alias{sf_writeLines}
\title{sf_writeLines}
\usage{
sf_writeLines(text, file, sep = "\n", na_value = "NA", encode_mode = "UTF-8")
}
\arguments{
\item{text}{A character to write to file}

\item{file}{Name of the file to write to}

\item{sep}{The line separator character(s)}

\item{na_value}{What to write in case of a NA string}

\item{encode_mode}{"UTF-8" or "byte". If "UTF-8", text strings are normalized to UTF-8 while `CE_BYTES` strings are written as raw bytes.}
}
\description{
A function that writes text line by line
}
\details{
A function for writing text data using `std::ofstream`.
}
\examples{
file <- tempfile()
sf_writeLines(letters, file)
sf_readLines(file)
}
\seealso{
writeLines
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zz_help_files.R
\name{string_identical}
\alias{string_identical}
\title{string_identical}
\usage{
string_identical(x, y, mode = c("semantic", "exact"))
}
\arguments{
\item{x}{A character vector}

\item{y}{Another character vector to compare to x}

\item{mode}{Either `"semantic"` to compare text after normalizing non-byte
strings to UTF-8, or `"exact"` to additionally require matching encoding.
Strings marked as `"bytes"` are always compared exactly.}
}
\value{
TRUE if strings are identical under the selected comparison mode
}
\description{
Compare strings semantically or exactly
}
\examples{
x <- "fa\xE7ile"
Encoding(x) <- "latin1"
y <- iconv(x, "latin1", "UTF-8")
identical(x, y) # TRUE
string_identical(x, y) # TRUE
string_identical(x, y, mode = "exact") # FALSE
}
\seealso{
identical
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zz_help_files.R
\name{sf_ends}
\alias{sf_ends}
\title{sf_ends}
\usage{
sf_ends(subject, pattern, ...)
}
\arguments{
\item{subject}{A character vector}

\item{pattern}{A string to look for at the start}

\item{...}{Parameters passed to sf_grepl}
}
\value{
A logical vector true if there is a match, false if no match, NA is the subject was NA
}
\description{
A function for detecting a pattern at the end of a string
}
\examples{
x <- c("alpha", "beta", "gamma", "delta", "epsilon")
sf_ends(x, "a")
}
\seealso{
endsWith, sf_starts
}
